"""Interface classes for `pyglet.input`.

.. versionadded:: 1.2
"""
from __future__ import annotations

import sys
import enum
import warnings

from typing import TYPE_CHECKING, Literal

from pyglet.math import Vec2
from pyglet.event import EventDispatcher

if TYPE_CHECKING:
    from pyglet.window import BaseWindow
    from pyglet.display.base import Display
    from pyglet.input.controller import Relation

_is_pyglet_doc_run = hasattr(sys, "is_pyglet_doc_run") and sys.is_pyglet_doc_run


class DeviceException(Exception):
    pass


class DeviceOpenException(DeviceException):
    pass


class DeviceExclusiveException(DeviceException):
    pass


class Sign(enum.Enum):
    POSITIVE = enum.auto()
    NEGATIVE = enum.auto()
    INVERTED = enum.auto()
    DEFAULT = enum.auto()


class Device:
    """Low level input device."""

    connected: bool

    def __init__(self, display: Display, name: str) -> None:
        """Create a Device to receive input from.

        Args:
            display:
                The Display this device is connected to.
            name:
                The name of the device, as described by the device firmware.
        """
        self.display = display
        self.name = name

        #: The manufacturer name, if available
        self.manufacturer: str | None = None
        self._is_open: bool = False

    @property
    def is_open(self) -> bool:
        return self._is_open

    def open(self, window: None | BaseWindow = None, exclusive: bool = False) -> None:
        """Open the device to begin receiving input from it.

        Args:
            window:
                Optional window to associate with the device.  The behaviour
                of this parameter is device and operating system dependent.
                It can usually be omitted for most devices.
            exclusive:
                If ``True`` the device will be opened exclusively so that no
                other application can use it.

        Raises:
            DeviceOpenException:
                If the device cannot be opened in exclusive mode, usually
                due to being opened exclusively by another application.
        """

        if self._is_open:
            raise DeviceOpenException('Device is already open.')

        self._is_open = True

    def close(self) -> None:
        """Close the device."""
        self._is_open = False

    def get_controls(self) -> list[Control]:
        """Get a list of controls provided by the device."""
        raise NotImplementedError('abstract')

    def get_guid(self) -> str:
        """Get the device GUID, in SDL2 format.

        Return a str containing a unique device identification
        string. This is generated from the hardware identifiers,
        and is in the same format as was popularized by SDL2.
        GUIDs differ between platforms, but are generally 32
        hexadecimal characters.
        """
        raise NotImplementedError('abstract')

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name})"


class Control(EventDispatcher):
    """Single value input provided by a device.

    A control's value can be queried when the device is open.  Event handlers
    can be attached to the control to be called when the value changes.

    The ``min`` and ``max`` properties are provided as advertised by the
    device; in some cases the control's value will be outside this range.
    """

    def __init__(self, name: None | str, raw_name: None | str = None, inverted: bool = False):
        """Create a Control to receive input.

        Args:
            name:
                The name of the control, or ``None`` if unknown.
            raw_name:
                Optional unmodified name of the control, as presented by the operating
                system; or ``None`` if unknown.
            inverted:
                If ``True``, the value reported is actually inverted from what the
                device reported; usually this is to provide consistency across
                operating systems.
        """
        self.name = name
        self.raw_name = raw_name
        self.inverted = inverted
        self._value = None

    @property
    def value(self) -> float:
        """Current value of the control.

        The range of the value is device-dependent; for absolute controls
        the range is given by ``min`` and ``max`` (however the value may exceed
        this range); for relative controls the range is undefined.
        """
        return self._value

    @value.setter
    def value(self, newvalue: float):
        if newvalue == self._value:
            return
        self._value = newvalue
        self.dispatch_event('on_change', newvalue)

    def __repr__(self) -> str:
        if self.name:
            return f"{self.__class__.__name__}(name={self.name}, raw_name={self.raw_name})"
        else:
            return f"{self.__class__.__name__}(raw_name={self.raw_name})"

    # Events

    def on_change(self, value) -> float:
        """The value changed."""


Control.register_event_type('on_change')


class RelativeAxis(Control):
    """An axis whose value represents a relative change from the previous value.

    This type of axis is used for controls that can scroll or move
    continuously, such as a scrolling or pointing input. The value
    is read as a delta change from the previous value.
    """
    X: str = 'x'
    Y: str = 'y'
    Z: str = 'z'
    RX: str = 'rx'
    RY: str = 'ry'
    RZ: str = 'rz'
    WHEEL: str = 'wheel'

    @property
    def value(self) -> float:
        return self._value

    @value.setter
    def value(self, value: float):
        self._value = value
        self.dispatch_event('on_change', value)


class AbsoluteAxis(Control):
    """An axis whose value represents the current measurement from the device.

    This type of axis is used for controls that have minimum and maximum
    positions. The value is a range between the ``min`` and ``max``.
    """
    X: str = 'x'
    Y: str = 'y'
    Z: str = 'z'
    RX: str = 'rx'
    RY: str = 'ry'
    RZ: str = 'rz'
    HAT: str = 'hat'
    HAT_X: str = 'hat_x'
    HAT_Y: str = 'hat_y'

    def __init__(self, name: str, minimum: float, maximum: float, raw_name: None | str = None, inverted: bool = False):
        super().__init__(name, raw_name, inverted)
        self.min = minimum
        self.max = maximum


class Button(Control):
    """A control whose value is boolean. """

    @property
    def value(self) -> bool:
        return bool(self._value)

    @value.setter
    def value(self, newvalue: bool | int):
        if newvalue == self._value:
            return
        self._value = newvalue
        self.dispatch_event('on_change', bool(newvalue))
        if newvalue:
            self.dispatch_event('on_press')
        else:
            self.dispatch_event('on_release')

    if _is_pyglet_doc_run:
        # Events

        def on_press(self):
            """The button was pressed."""

        def on_release(self):
            """The button was released."""


Button.register_event_type('on_press')
Button.register_event_type('on_release')


class Joystick(EventDispatcher):
    """High-level interface for joystick-like devices.  This includes a wide
    range of analog and digital joysticks, gamepads, controllers, and possibly
    even steering wheels and other input devices. There is unfortunately no
    easy way to distinguish between most of these different device types.

    For a simplified subset of Joysticks, see the :py:class:`~pyglet.input.Controller`
    interface. This covers a variety of popular game console controllers. Unlike
    Joysticks, Controllers have strictly defined layouts and inputs.

    To use a joystick, first call `open`, then in your game loop examine
    the values of `x`, `y`, and so on.  These values are normalized to the
    range [-1.0, 1.0]. 

    To receive events when the value of an axis changes, attach an
    on_joyaxis_motion event handler to the joystick. The :py:class:`~pyglet.input.Joystick`
    instance, axis name, and current value are passed as parameters to this event.

    To handle button events, you should attach on_joybutton_press and on_joy_button_release
    event handlers to the joystick. Both the :py:class:`~pyglet.input.Joystick` instance
    and the index of the changed button are passed as parameters to these events.

    Alternately, you may attach event handlers to each individual button in 
    `button_controls` to receive on_press or on_release events.
    
    To use the hat switch, attach an on_joyhat_motion event handler to the joystick.
    The handler will be called with both the hat_x and hat_y values whenever the value
    of the hat switch changes.

    The device name can be queried to get the name of the joystick.
    """
    #: :The underlying device used by this joystick interface.
    device: Device

    #: :Current X (horizontal) value ranging from -1.0 (left) to 1.0 (right).
    x: float
    #: :Current y (vertical) value ranging from -1.0 (top) to 1.0 (bottom).
    y: float
    #: :Current Z value ranging from -1.0 to 1.0.  On joysticks the Z value is usually the
    #: :throttle control. On controllers the Z value is usually the secondary thumb vertical axis.
    z: float

    #: :Current rotational X value ranging from -1.0 to 1.0.
    rx: float
    #: :Current rotational Y value ranging from -1.0 to 1.0.
    ry: float
    #: :Current rotational Z value ranging from -1.0 to 1.0.  On joysticks the RZ value
    #: :is usually the twist of the stick.  On game controllers the RZ value is usually
    #: :the secondary thumb horizontal axis.
    rz: float

    #: :Current hat (POV) horizontal position; one of -1 (left), 0 (centered) or 1 (right).
    hat_x: int
    #: :Current hat (POV) vertical position; one of -1 (bottom), 0 (centered) or 1 (top).
    hat_y: int

    #: :List of boolean values representing current states of the buttons. These
    #: :are in order, so that button 1 has value at ``buttons[0]``, and so on.
    buttons: list[bool]

    #: :Underlying control for ``x`` value, or ``None`` if not available.
    x_control: None | AbsoluteAxis
    #: :Underlying control for ``y`` value, or ``None`` if not available.
    y_control: None | AbsoluteAxis
    #: :Underlying control for ``z`` value, or ``None`` if not available.
    z_control: None | AbsoluteAxis

    #: :Underlying control for ``rx`` value, or ``None`` if not available.
    rx_control: None | AbsoluteAxis
    #: :Underlying control for ``ry`` value, or ``None`` if not available.
    ry_control: None | AbsoluteAxis
    #: :Underlying control for ``rz`` value, or ``None`` if not available.
    rz_control: None | AbsoluteAxis

    #: :Underlying control for ``hat_x`` value, or ``None`` if not available.
    hat_x_control: None | AbsoluteAxis
    #: :Underlying control for ``hat_y`` value, or ``None`` if not available.
    hat_y_control: None | AbsoluteAxis

    #: Underlying controls for ``buttons`` values.
    button_controls: list[Button]

    def __init__(self, device):
        self.device = device

        self.x = 0
        self.y = 0
        self.z = 0
        self.rx = 0
        self.ry = 0
        self.rz = 0
        self.hat_x = 0
        self.hat_y = 0
        self.buttons = []

        self.x_control = None
        self.y_control = None
        self.z_control = None
        self.rx_control = None
        self.ry_control = None
        self.rz_control = None
        self.hat_x_control = None
        self.hat_y_control = None
        self.button_controls = []

        def add_axis(control: AbsoluteAxis):
            if not (control.min or control.max):
                warnings.warn(f"Control('{control.name}') min & max values are both 0. Skipping.")
                return
            name = control.name
            scale = 2.0 / (control.max - control.min)
            bias = -1.0 - control.min * scale
            if control.inverted:
                scale = -scale
                bias = -bias
            setattr(self, name + '_control', control)

            @control.event
            def on_change(value):
                normalized_value = value * scale + bias
                setattr(self, name, normalized_value)
                self.dispatch_event('on_joyaxis_motion', self, name, normalized_value)

        def add_button(control: Button):
            i = len(self.buttons)
            self.buttons.append(False)
            self.button_controls.append(control)

            @control.event
            def on_change(value):
                self.buttons[i] = value

            @control.event
            def on_press():
                self.dispatch_event('on_joybutton_press', self, i)

            @control.event
            def on_release():
                self.dispatch_event('on_joybutton_release', self, i)

        def add_hat(control: AbsoluteAxis):
            # 8-directional hat encoded as a single control (Windows/Mac)
            self.hat_x_control = control
            self.hat_y_control = control

            _vecs = (Vec2(0.0, 1.0), Vec2(1.0, 1.0), Vec2(1.0, 0.0), Vec2(1.0, -1.0),  # n, ne, e, se
                     Vec2(0.0, -1.0), Vec2(-1.0, -1.0), Vec2(-1.0, 0.0), Vec2(-1.0, 1.0))  # s, sw, w, nw
            _input_map = {key: val for key, val in zip(range(int(control.min), int(control.max + 1)), _vecs)}

            @control.event
            def on_change(value):
                self.hat_x, self.hat_y = _input_map.get(value, Vec2(0.0, 0.0))
                self.dispatch_event('on_joyhat_motion', self, self.hat_x, self.hat_y)

        for ctrl in device.get_controls():
            if isinstance(ctrl, AbsoluteAxis):
                if ctrl.name in ('x', 'y', 'z', 'rx', 'ry', 'rz', 'hat_x', 'hat_y'):
                    add_axis(ctrl)
                elif ctrl.name == 'hat':
                    add_hat(ctrl)
            elif isinstance(ctrl, Button):
                add_button(ctrl)

    def open(self, window: BaseWindow | None = None, exclusive: bool = False) -> None:
        """Open the joystick device.  See `Device.open`. """
        self.device.open(window, exclusive)

    def close(self) -> None:
        """Close the joystick device.  See `Device.close`. """
        self.device.close()

    # Events

    def on_joyaxis_motion(self, joystick: Joystick, axis: str, value: float):
        """The value of a joystick axis changed.

        Args:
            joystick:
                The joystick device whose axis changed.
            axis:
                The name of the axis that changed.
            value:
                The current value of the axis, normalized to [-1, 1].
        """

    def on_joybutton_press(self, joystick: Joystick, button: int):
        """A button on the joystick was pressed.

        Args:
            joystick:
                The joystick device whose button was pressed.
            button:
                The index (in `button_controls`) of the button that was pressed.
        """

    def on_joybutton_release(self, joystick: Joystick, button: int):
        """A button on the joystick was released.

        Args:
            joystick:
                The joystick device whose button was released.
            button:
                The index (in `button_controls`) of the button that was released.
        """

    def on_joyhat_motion(self, joystick: Joystick, hat_x: float, hat_y: float):
        """The value of the joystick hat switch changed.

        Args:
            joystick:
                The joystick device whose hat control changed.
            hat_x:
                Current hat (POV) horizontal position; one of -1.0 (left), 0.0
                (centered) or 1.0 (right).
            hat_y:
                Current hat (POV) vertical position; one of -1.0 (bottom), 0.0
                (centered) or 1.0 (top).
        """

    def __repr__(self) -> str:
        return f"Joystick(device={self.device.name})"


Joystick.register_event_type('on_joyaxis_motion')
Joystick.register_event_type('on_joybutton_press')
Joystick.register_event_type('on_joybutton_release')
Joystick.register_event_type('on_joyhat_motion')


class Controller(EventDispatcher):
    """High-level interface for Game Controllers.

    Unlike Joysticks, Controllers have a strictly defined set of inputs
    that matches the layout of popular home video game console Controllers.
    This includes a variety of face and shoulder buttons, analog sticks and
    triggers, a directional pad, and optional rumble (force feedback) effects.

    To use a Controller, you must first call ``open``. Controllers will then
    dispatch various events whenever the inputs change. They can also be polled
    manually at any time to find the current value of any inputs. Analog stick
    inputs are normalized to the range [-1.0, 1.0], and triggers are normalized
    to the range [0.0, 1.0]. All other inputs are digital.

    Note: A running application event loop is required

    The following event types are dispatched:
        `on_button_press`
        `on_button_release`
        `on_stick_motion`
        `on_dpad_motion`
        `on_trigger_motion`

    """

    def __init__(self, device: Device, mapping: dict):
        """Create a Controller instance mapped to a Device.

        .. versionadded:: 2.0
        """

        #: The underlying Device:
        self.device: Device = device
        self._mapping = mapping

        #: The logical device name
        self.name: str = mapping.get('name')
        #: The unique guid for this Device
        self.guid: str = mapping.get('guid')

        self.a: bool = False
        self.b: bool = False
        self.x: bool = False
        self.y: bool = False
        self.back: bool = False
        self.start: bool = False
        self.guide: bool = False
        self.leftshoulder: bool = False
        self.rightshoulder: bool = False
        self.leftstick: bool = False          # stick press button
        self.rightstick: bool = False         # stick press button

        self.lefttrigger: float = 0.0
        self.righttrigger: float = 0.0
        self.leftx: float = 0.0
        self.lefty: float = 0.0
        self.rightx: float = 0.0
        self.righty: float = 0.0
        self.dpadx: float = 0.0
        self.dpady: float = 0.0

        self._button_controls: list = []
        self._axis_controls: list = []
        self._hat_control: Control | None = None
        self._hat_x_control: Control | None = None
        self._hat_y_control: Control | None = None

        self._initialize_controls()

    @property
    def type(self) -> Literal['PS', 'XB', 'GENERIC']:
        """The type, or family, of the Controller

        This property uses a simple heuristic query to attempt
        to determine which general family the controller falls
        into. For example, the controller may have Ⓐ,Ⓑ,Ⓧ,Ⓨ,
        or ✕,○,□,△ labels on the face buttons. Using this
        information, you can choose to display matching button
        prompt images in your game. For example::

            if controller.type == 'PS':
                a_glyph = 'ps_cross_button.png'
                b_glyph = 'ps_circle_button.png'
                ...
            elif controller.type == 'XB':
                a_glyph = 'ms_a_button.png'
                b_glyph = 'ms_b_button.png'
                ...
            else:
                ...

        Returns:
            A string, currently one of "PS", "XB", or "GENERIC".
        """
        product_id = None
        # TODO: add more checks for vendor hardware ids.

        # Windows
        if self.name == 'XINPUTCONTROLLER':
            return 'XB'

        # Linux
        if id_product := getattr(self.device, 'id_product'):
            product_id = int(id_product, base=0)

        if product_id in (0x5C4,):
            return 'PS'
        elif product_id in (0x045E,):
            return 'XB'

        return 'GENERIC'

    def _bind_axis_control(self, relation: Relation, control: AbsoluteAxis, axis_name: str) -> None:
        if not (control.min or control.max):
            warnings.warn(f"Control('{control.name}') min & max values are both 0. Skipping.")
            return

        tscale = 1.0 / (control.max - control.min)
        scale = 2.0 / (control.max - control.min)
        bias = -1.0 - control.min * scale
        sign = 1.0
        if control.inverted:
            scale = -scale
            bias = -bias

        # Default Signs, if not explicitly defined in the mapping string:
        dpad_defaults = {'dpup': Sign.POSITIVE, 'dpdown': Sign.NEGATIVE,
                         'dpleft': Sign.NEGATIVE, 'dpright': Sign.POSITIVE}

        # If the sign is not DEFAULT, it must be inverted:
        if relation.sign not in (Sign.DEFAULT, dpad_defaults.get(axis_name)):
            sign = -1.0

        if axis_name in ("dpup", "dpdown"):
            @control.event
            def on_change(value):
                self.dpady = round(value * scale + bias) * sign    # normalized
                self.dispatch_event('on_dpad_motion', self, Vec2(self.dpadx, self.dpady))

        elif axis_name in ("dpleft", "dpright"):
            @control.event
            def on_change(value):
                self.dpadx = round(value * scale + bias) * sign     # normalized
                self.dispatch_event('on_dpad_motion', self, Vec2(self.dpadx, self.dpady))

        elif axis_name in ("lefttrigger", "righttrigger"):
            @control.event
            def on_change(value):
                normalized_value = value * tscale
                setattr(self, axis_name, normalized_value)
                self.dispatch_event('on_trigger_motion', self, axis_name, normalized_value)

        elif axis_name in ("leftx", "lefty"):
            @control.event
            def on_change(value):
                normalized_value = value * scale + bias
                setattr(self, axis_name, normalized_value)
                self.dispatch_event('on_stick_motion', self, "leftstick", Vec2(self.leftx, -self.lefty))

        elif axis_name in ("rightx", "righty"):
            @control.event
            def on_change(value):
                normalized_value = value * scale + bias
                setattr(self, axis_name, normalized_value)
                self.dispatch_event('on_stick_motion', self, "rightstick", Vec2(self.rightx, -self.righty))

    def _bind_button_control(self, relation: Relation, control: Button, button_name: str) -> None:
        if button_name in ("dpleft", "dpright", "dpup", "dpdown"):
            defaults = {'dpleft': ('dpadx', -1.0), 'dpright': ('dpadx', 1.0),
                        'dpdown': ('dpady', -1.0), 'dpup': ('dpady', 1.0)}

            @control.event
            def on_change(value):
                target, bias = defaults[button_name]
                setattr(self, target, bias * value)
                self.dispatch_event('on_dpad_motion', self, Vec2(self.dpadx, self.dpady))
        else:
            @control.event
            def on_change(value):
                setattr(self, button_name, value)

            @control.event
            def on_press():
                self.dispatch_event('on_button_press', self, button_name)

            @control.event
            def on_release():
                self.dispatch_event('on_button_release', self, button_name)

    def _bind_dedicated_hat(self, relation: Relation, control: AbsoluteAxis) -> None:
        # 8-directional hat encoded as a single control (Windows/Mac)
        _vecs = (Vec2(0.0, 1.0), Vec2(1.0, 1.0), Vec2(1.0, 0.0), Vec2(1.0, -1.0),       # n, ne, e, se
                 Vec2(0.0, -1.0), Vec2(-1.0, -1.0), Vec2(-1.0, 0.0), Vec2(-1.0, 1.0))   # s, sw, w, nw
        _input_map = {key: val for key, val in zip(range(int(control.min), int(control.max + 1)), _vecs)}

        # For some Directinput devices:
        _scale = 0xfff if control.max > 8 else 1

        @control.event
        def on_change(value):
            vector = _input_map.get(value // _scale, Vec2(0.0, 0.0))
            self.dpadx, self.dpady = vector
            self.dispatch_event('on_dpad_motion', self, vector)

    def _initialize_controls(self) -> None:
        """Initialize and bind the Device Controls

        This method first categorizes all the Device Controls,
        then binds them to the appropriate "virtual" controls
        as defined in the mapped relations.
        """

        for ctrl in self.device.get_controls():
            # Categorize the various control types
            if isinstance(ctrl, Button):
                self._button_controls.append(ctrl)

            elif isinstance(ctrl, AbsoluteAxis):
                if ctrl.name == "hat_x":
                    self._hat_x_control = ctrl
                elif ctrl.name == "hat_y":
                    self._hat_y_control = ctrl
                elif ctrl.name == "hat":
                    self._hat_control = ctrl
                else:
                    self._axis_controls.append(ctrl)

        for name, relation in self._mapping.items():

            if relation is None or isinstance(relation, str):
                continue

            try:
                if relation.control_type == "button":
                    self._bind_button_control(relation, self._button_controls[relation.index], name)

                elif relation.control_type == "axis":
                    self._bind_axis_control(relation, self._axis_controls[relation.index], name)

                elif relation.control_type == "hat0":
                    if self._hat_control:
                        self._bind_dedicated_hat(relation, self._hat_control)
                    else:
                        control, dpname = {1: (self._hat_y_control, 'dpup'),
                                           2: (self._hat_x_control, 'dpright'),
                                           4: (self._hat_y_control, 'dpdown'),
                                           8: (self._hat_x_control, 'dpleft')}[relation.index]

                        self._bind_axis_control(relation, control, dpname)

            except (IndexError, AttributeError, KeyError):
                warnings.warn(f"Could not map physical Control '{relation}' to '{name}'")
                continue

    def open(self, window: None | BaseWindow = None, exclusive: bool = False) -> None:
        """Open the controller.  See `Device.open`. """
        self.device.open(window, exclusive)

    def close(self) -> None:
        """Close the controller.  See `Device.close`. """
        self.device.close()

    def rumble_play_weak(self, strength: float = 1.0, duration: float = 0.5) -> None:
        """Play a rumble effect on the weak motor.

        Args:
            strength:
                The strength of the effect, from 0 to 1.
            duration:
                The duration of the effect in seconds.
        """

    def rumble_play_strong(self, strength: float = 1.0, duration: float = 0.5) -> None:
        """Play a rumble effect on the strong motor.

        Args:
            strength:
                The strength of the effect, from 0 to 1.
            duration:
                The duration of the effect in seconds.
        """

    def rumble_stop_weak(self) -> None:
        """Stop playing rumble effects on the weak motor."""

    def rumble_stop_strong(self) -> None:
        """Stop playing rumble effects on the strong motor."""

    # Events

    def on_stick_motion(self, controller: Controller, stick: str, vector: Vec2):
        """The value of a controller analogue stick changed.

        Args:
            controller:
                The controller whose analogue stick changed.
            stick:
                The name of the stick that changed.
            vector:
                A 2D vector representing the stick position.
                Each individual axis will be normalized from [-1, 1],
        """

    def on_dpad_motion(self, controller: Controller, vector: Vec2):
        """The direction pad of the controller changed.

        Args:
            controller:
                The controller whose hat control changed.
            vector:
                A 2D vector, representing the dpad position.
                Each individual axis will be one of [-1, 0, 1].
        """

    def on_trigger_motion(self, controller: Controller, trigger: str, value: float):
        """The value of a controller analogue stick changed.

        Args:
            controller:
                The controller whose analogue stick changed.
            trigger:
                The name of the trigger that changed.
            value:
                The current value of the trigger, normalized to [0, 1].
        """

    def on_button_press(self, controller: Controller, button: str):
        """A button on the controller was pressed.

        Args:
            controller:
                The controller whose button was pressed.
            button:
                The name of the button that was pressed.
        """

    def on_button_release(self, controller: Controller, button: str):
        """A button on the joystick was released.

        Args:
            controller:
                The controller whose button was released.
            button:
                The name of the button that was released.
        """

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name})"


Controller.register_event_type('on_button_press')
Controller.register_event_type('on_button_release')
Controller.register_event_type('on_stick_motion')
Controller.register_event_type('on_dpad_motion')
Controller.register_event_type('on_trigger_motion')


class AppleRemote(EventDispatcher):
    """High-level interface for Apple remote control.

    This interface provides access to the 6 button controls on the remote.
    Pressing and holding certain buttons on the remote is interpreted as
    a separate control.
    """

    device: Device
    left_control: Button
    left_hold_control: Button
    right_control: Button
    right_hold_control: Button
    up_control: Button
    down_control: Button
    select_control: Button
    select_hold_control: Button
    menu_control: Button
    menu_hold_control: Button

    def __init__(self, device):
        self.device = device

        def _add_button(button: Button):
            setattr(self, button.name + '_control', button)

            @button.event
            def on_press():
                self.dispatch_event('on_button_press', button.name)

            @button.event
            def on_release():
                self.dispatch_event('on_button_release', button.name)

        for control in device.get_controls():
            if control.name in ('left', 'left_hold', 'right', 'right_hold', 'up', 'down',
                                'menu', 'select', 'menu_hold', 'select_hold'):
                _add_button(control)

    def open(self, window: BaseWindow, exclusive: bool = False):
        """Open the device.  See `Device.open`. """
        self.device.open(window, exclusive)

    def close(self):
        """Close the device.  See `Device.close`. """
        self.device.close()

    # Events

    def on_button_press(self, button: str):
        """A button on the remote was pressed.

        Only the 'up' and 'down' buttons will generate an event when the
        button is first pressed.  All other buttons on the remote will wait
        until the button is released and then send both the press and release
        events at the same time.

        Args:
            button:
                The name of the button that was pressed. The valid names are
                'up', 'down', 'left', 'right', 'left_hold', 'right_hold',
                'menu', 'menu_hold', 'select', and 'select_hold'
        """

    def on_button_release(self, button: str):
        """A button on the remote was released.

        The 'select_hold' and 'menu_hold' button release events are sent
        immediately after the corresponding press events regardless of
        whether the user has released the button.

        Args:
            button:
                The name of the button that was released. The valid names are
                'up', 'down', 'left', 'right', 'left_hold', 'right_hold',
                'menu', 'menu_hold', 'select', and 'select_hold'
        """


AppleRemote.register_event_type('on_button_press')
AppleRemote.register_event_type('on_button_release')


class Tablet:
    """High-level interface to tablet devices.

    Unlike other devices, tablets must be opened for a specific window,
    and cannot be opened exclusively.  The `open` method returns a
    `TabletCanvas` object, which supports the events provided by the tablet.

    Currently only one tablet device can be used, though it can be opened on
    multiple windows.  If more than one tablet is connected, the behaviour is
    undefined.
    """

    def open(self, window: BaseWindow) -> TabletCanvas:
        """Open a tablet device for a window.

        Args:
            window:
                The window on which the tablet will be used.
        """
        raise NotImplementedError('abstract')


class TabletCanvas(EventDispatcher):
    """Event dispatcher for tablets.

    Use `Tablet.open` to obtain this object for a particular tablet device and
    window.  Events may be generated even if the tablet stylus is outside of
    the window; this is operating-system dependent.

    The events each provide the `TabletCursor` that was used to generate the
    event; for example, to distinguish between a stylus and an eraser.  Only
    one cursor can be used at a time, otherwise the results are undefined.
    """

    # OS X: Active window receives tablet events only when cursor is in window
    # Windows: Active window receives all tablet events
    #
    # Note that this means enter/leave pairs are not always consistent (normal
    # usage).

    def __init__(self, window: BaseWindow):
        """Create a TabletCanvas.

        Args:
            window:
                The window on which this tablet was opened.
        """
        self.window = window

    def close(self) -> None:
        """Close the tablet device for this window."""
        raise NotImplementedError('abstract')

    if _is_pyglet_doc_run:
        # Events

        def on_enter(self, cursor: TabletCursor):
            """A cursor entered the proximity of the window.  The cursor may
            be hovering above the tablet surface, but outside the window
            bounds, or it may have entered the window bounds.

            Note that you cannot rely on `on_enter` and `on_leave` events to
            be generated in pairs; some events may be lost if the cursor was
            out of the window bounds at the time.

            :event:
            """

        def on_leave(self, cursor: TabletCursor):
            """A cursor left the proximity of the window.  The cursor may have
            moved too high above the tablet surface to be detected, or it may
            have left the bounds of the window.

            Note that you cannot rely on `on_enter` and `on_leave` events to
            be generated in pairs; some events may be lost if the cursor was
            out of the window bounds at the time.
            """

        def on_motion(self, cursor: TabletCursor, x: int, y: int,
                      pressure: float, tilt_x: float, tilt_y: float, buttons: int):
            """The cursor moved on the tablet surface.

            If `pressure` is 0, then the cursor is actually hovering above the
            tablet surface, not in contact.

            Args:
                cursor:
                    The cursor that moved.
                x:
                    The X position of the cursor, in window coordinates.
                y:
                    The Y position of the cursor, in window coordinates.
                pressure:
                    The pressure applied to the cursor, in range 0.0
                    (no pressure) to 1.0 (full pressure).
                tilt_x:
                    Currently undefined.
                tilt_y:
                    Currently undefined.
                buttons:
                    Button state may be provided if the platform supports it.
                    Supported on: Windows
            """


TabletCanvas.register_event_type('on_enter')
TabletCanvas.register_event_type('on_leave')
TabletCanvas.register_event_type('on_motion')


class TabletCursor:
    """A distinct cursor used on a Tablet device.

    Most tablets support at least a ``stylus`` and an ``erasor`` cursor;
    this object is used to distinguish them when tablet events are generated.
    """

    # TODO well-defined names for stylus and eraser.

    def __init__(self, name: str):
        """Create a cursor object.

        Args:
            name:
                The name of the cursor.
        """
        self.name = name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name})"


class ControllerManager(EventDispatcher):
    """High level interface for managing game Controllers.

    This class provides a convenient way to handle the
    connection and disconnection of devices. A list of all
    connected Controllers can be queried at any time with the
    `get_controllers` method. For hot-plugging, events are
    dispatched for `on_connect` and `on_disconnect`.
    To use the ControllerManager, first make an instance::

        controller_man = pyglet.input.ControllerManager()

    At the start of your game, query for any Controllers
    that are already connected::

        controllers = controller_man.get_controllers()

    To handle Controllers that are connected or disconnected
    after the start of your game, register handlers for the
    appropriate events::

        @controller_man.event
        def on_connect(controller):
            # code to handle newly connected
            # (or re-connected) controllers
            controller.open()
            print("Connect:", controller)

        @controller_man.event
        def on_disconnect(controller):
            # code to handle disconnected Controller
            print("Disconnect:", controller)

    .. versionadded:: 2.0
    """

    def get_controllers(self) -> list[Controller]:
        """Get a list of all connected Controllers"""
        raise NotImplementedError

    # Events

    def on_connect(self, controller) -> Controller:
        """A Controller has been connected. If this is
        a previously disconnected Controller that is
        being re-connected, the same Controller instance
        will be returned.
        """

    def on_disconnect(self, controller) -> Controller:
        """A Controller has been disconnected."""


ControllerManager.register_event_type('on_connect')
ControllerManager.register_event_type('on_disconnect')



"""An example sprite previewer using Qt and pyglet together.

It allows you to edit the fragment and vertex shaders, then compile them
to get a live view. Errors and success will be printed to the console. Users
running GNOME on Linux may need the --use-qt-file-dialog flag to prevent invisible
window issues. A brief feature overview is located after the license notes below.

Important license notes:

1. Libraries can use licenses which impose requirements beyond those of
   pyglet's BSD-style style license.
2. This example defaults to using PySide6 by default, but can also use
   PyQt6 due to their nearly-identical APIs.
3. PySide6 uses the LGPL license while PyQt6 uses a GPL / commercial
   dual-license approach.

To the best knowledge of the contributors, this example and derivatives
are only obligated to meet the restrictions of the LGPL because it does
not use any PyQt6-specific features. Please see the following for more
information:

* The licenses and documentation for the libraries you plan to use
* This guide for prior versions of PySide and PyQt:
  thttps://www.pythonguis.com/faq/pyqt5-vs-pyside2/

For additional certainty, please consult a legal professional.


Example features:

You can choose the current Qt binding in two ways:

1. Add PySide6 or PyQt6 as the first argument after the launch command
   when running the script in the terminal
2. Set the PYGLET_QT_BACKEND environment variable to either PySide6 or
   PyQt6.

The priority order is:

1. positional argument
2. environment variable
3. default to PySide6

To load images, choose File -> Open Image.

Images loaded will be listed in the Images menu. By selecting an image in
the menu list, it will be unloaded. Names in parentheses will be used for
the sampler2D name.

You can open a shader (both vert and frag) at the same time. Text is
allowed, but will load into the fragment shader. Saving a shader saves
both frag and vertex (IE: test becomes test.vert and test.frag)

Scrolling the mousewheel also zooms in.
"""
from __future__ import annotations

import argparse
import os
import sys
import traceback
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Final, Mapping

# Constants for choosing a Qt backend and summarizing their licensing
ENV_VARIABLE: Final[str] = 'PYGLET_QT_BACKEND'

PYSIDE6: Final[str] = 'PySide6'
PYQT6: Final[str] = 'PyQt6'

valid_backends: Final[dict[str, str]] = {
    PYSIDE6: "LGPL; may allow releasing under non-GPL licenses",
    PYQT6: "Dual GPL / Commercial license; requires GPL or a fee",
}

# Default to PySide6 as part of allowing non-GPL licenses
# This is necessary but not sufficient to allow this.
DEFAULT: Final[str] = PYSIDE6


# Argument parser for run-time config as __main__
parser = argparse.ArgumentParser(
    description=dedent("""\
    A sprite & shader previewer using Qt and pyglet together.

    It defaults to PySide6, but can run on PyQt6. The details of how may
    be helpful to users who want to avoid spreading PyQt6's GPL license
    while using it as a fallback. See the docstrings and comments in the
    source to learn more.
    """),
    formatter_class=argparse.RawTextHelpFormatter)

# On Python 3.9+, argparse.BooleanOptionalAction is more concise
# See https://docs.python.org/3.9/library/argparse.html#action
parser.add_argument(
    '--use-qt-file-dialog', dest='native_file_dialog', action='store_false',
    help="Use Qt's file chooser instead of the system's. Helpful on Gnome DE.",
)
parser.set_defaults(native_file_dialog=True)


# Generate options help text lines
backend_column_width: Final[int] = max(map(len, valid_backends))
help_lines: list[str] = ["Which Qt binding to use.\n"]

for backend, description in valid_backends.items():
   line_parts = [
       f"{backend: <{backend_column_width}} - ",
       '(Default) ' if backend == DEFAULT else '',
       description,
   ]
   help_lines.append(''.join(line_parts))

# Add an optional single-value positional argument
parser.add_argument('backend', nargs='?',
    choices=valid_backends.keys(),
    help='\n'.join(help_lines))


# Use the 1st non-None Qt binding value: argument, env var, or default
arguments = None
if __name__ == "__main__":
    arguments = parser.parse_args()

reasons_with_backend_values: Final[dict[str, str | None]] = {
   'first positional argument'       : None if not arguments else arguments.backend,
   # An environment variable allows specifying the binding when
   # importing the module instead of running it as a main program.
   f'{ENV_VARIABLE!r} env variable'  : os.environ.get(ENV_VARIABLE, None),
   'default'                         : DEFAULT,
}
reason, backend = next(filter(
    lambda t: t[1] is not None, reasons_with_backend_values.items()))

print(f"Selected {backend} as the Qt binding from the {reason}'s value.")


# Perform UI imports according to the detected configuration
import pyglet

# Use PySide6 for type checking, static analysis, tests, and linting
# to help avoid infection by PyQt6's GPL license since tools will mark
# uses of PyQt6-specific features as missing. For example, pyright is
# one of the strict type checking and linting tools which can help.
if TYPE_CHECKING or backend == PYSIDE6:
    from PySide6 import QtCore, QtWidgets, QtGui
    from PySide6.QtGui import QWheelEvent
    from PySide6.QtWidgets import QFileDialog
    from PySide6.QtOpenGLWidgets import QOpenGLWidget

elif backend == PYQT6:
    from PyQt6 import QtCore, QtWidgets, QtGui
    from PyQt6.QtGui import QWheelEvent
    from PyQt6.QtWidgets import QFileDialog
    from PyQt6.QtOpenGLWidgets import QOpenGLWidget

else:  # Handle import edge cases
    raise ValueError(
        f"Expected a value in {valid_backends},"
        f" but got {backend!r} via {reason}")

# Import the other constants after the UI libraries to avoid
# cluttering the symbol table when debugging import problems.
from pyglet.gl import (
   GL_BLEND,
   GL_COLOR_BUFFER_BIT,
   GL_DEPTH_BUFFER_BIT,
   GL_TEXTURE0,
   glActiveTexture,
   glBindTexture,
   glBlendFunc,
   glClear,
   glDisable,
   glEnable,
)

default_vertex_src = """#version 150 core
in vec3 translate;
in vec4 colors;
in vec3 tex_coords;
in vec2 scale;
in vec3 position;
in float rotation;

out vec4 vertex_colors;
out vec3 texture_coords;

uniform WindowBlock
{
    mat4 projection;
    mat4 view;
} window;

mat4 m_scale = mat4(1.0);
mat4 m_rotation = mat4(1.0);
mat4 m_translate = mat4(1.0);

void main()
{
    m_scale[0][0] = scale.x;
    m_scale[1][1] = scale.y;
    m_translate[3][0] = translate.x;
    m_translate[3][1] = translate.y;
    m_translate[3][2] = translate.z;
    m_rotation[0][0] =  cos(-radians(rotation)); 
    m_rotation[0][1] =  sin(-radians(rotation));
    m_rotation[1][0] = -sin(-radians(rotation));
    m_rotation[1][1] =  cos(-radians(rotation));

    gl_Position = window.projection * window.view * m_translate * m_rotation * m_scale * vec4(position, 1.0);

    vertex_colors = colors;
    texture_coords = tex_coords;
}
"""

default_frag_src = """#version 150 core
in vec4 vertex_colors;
in vec3 texture_coords;
out vec4 final_colors;

uniform sampler2D sprite_texture0;
uniform sampler2D sprite_texture1;
uniform float time;

void main()
{
    vec2 uv = texture_coords.xy;
    vec3 col = 0.5 + 0.5 * cos(time + uv.xyx + vec3(0, 2, 4));
    
    final_colors = texture(sprite_texture1, uv) * texture(sprite_texture0, uv) * vertex_colors * vec4(col, 1.0);
}"""


class MultiTextureSpriteGroup(pyglet.sprite.SpriteGroup):
    """A sprite group which uses multiple textures and samplers."""

    def __init__(
            self,
            textures: Mapping[str, pyglet.image.Texture],
            blend_src: int,
            blend_dest: int,
            program: pyglet.graphics.shader.ShaderProgram | None = None,
            parent: pyglet.graphics.Group | None = None,
    ) -> None:
        """Create a sprite group for multiple textures and samplers.

        All textures must share the same target type.

        :Parameters:
            `textures` :
                 A mapping of sampler names to texture data.
            `blend_src` :
                OpenGL blend source mode; for example,
                ``GL_SRC_ALPHA``.
            `blend_dest` :
                OpenGL blend destination mode; for example,
                ``GL_ONE_MINUS_SRC_ALPHA``.
            `parent` :
                Optional parent group.
        """
        self.images = textures
        texture = list(self.images.values())[0]
        self.target = texture.target
        super().__init__(texture, blend_src, blend_dest, program, parent)

        self.program.use()
        for idx, name in enumerate(self.images):
            try:
                self.program[name] = idx
            except pyglet.graphics.shader.ShaderException as e:
                print(e)

        self.program.stop()

    def set_state(self) -> None:
        self.program.use()

        for i, texture in enumerate(self.images.values()):
            glActiveTexture(GL_TEXTURE0 + i)
            glBindTexture(self.target, texture.id)

        glEnable(GL_BLEND)
        glBlendFunc(self.blend_src, self.blend_dest)

    def unset_state(self) -> None:
        glDisable(GL_BLEND)
        self.program.stop()
        glActiveTexture(GL_TEXTURE0)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.texture!r}-{int(self.texture.id)})'

    def __eq__(self, other) -> bool:
        return (other.__class__ is self.__class__ and
                self.program is other.program and
                self.images == other.textures and
                self.blend_src == other.blend_src and
                self.blend_dest == other.blend_dest)

    def __hash__(self) -> int:
        return hash((id(self.parent),
                     id(self.images),
                     self.blend_src, self.blend_dest))


class MultiTextureSprite(pyglet.sprite.Sprite):

    def __init__(
            self,
            imgs: Mapping[str, pyglet.image.Texture],
            x: float = 0, y: float = 0, z: float = 0,
            blend_src: int = pyglet.gl.GL_SRC_ALPHA,
            blend_dest: int = pyglet.gl.GL_ONE_MINUS_SRC_ALPHA,
            batch: pyglet.graphics.Batch | None = None,
            group: MultiTextureSpriteGroup | None = None,
            subpixel: bool = False,
            program: pyglet.graphics.shader.ShaderProgram = None,
    ) -> None:

        self._x = x
        self._y = y
        self._z = z

        self._texture = list(imgs.values())[0]

        if isinstance(self._texture, pyglet.image.TextureArrayRegion):
            self._program = program or pyglet.sprite.get_default_array_shader()
        else:
            self._program = program or pyglet.sprite.get_default_shader()

        self._batch = batch or pyglet.graphics.get_default_batch()

        self._group = MultiTextureSpriteGroup(imgs, blend_src, blend_dest, self.program, group)

        self._subpixel = subpixel
        self._create_vertex_list()


class Ui_MainWindow:
    SPRITE_POSITION = (0, 0)

    def __init__(self, use_native_file_dialog: bool = True):
        self.use_native_file_dialog: bool = use_native_file_dialog

        self.images = []

        self.group = pyglet.graphics.Group()

        self.sprite = None
        self.program = None

    def get_file_dialog_options(self) -> QFileDialog.Option:
        """Convert instance attributes to a file dialog options object.

        At the moment, it supports choosing which dialog to use. This is
        helpful on certain Gnome and tiling Linux desktop configurations
        which can have issues with the system file picker.

        You may want to expand on this in your own application with
        additional options.
        """
        options = QFileDialog.Option()
        if not self.use_native_file_dialog:
            options |= QFileDialog.Option.DontUseNativeDialog

        return options

    def setupUi(self, MainWindow: QtWidgets.QMainWindow) -> None:
        self._window = MainWindow

        # Set up the central window widget object
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(820, 855)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")

        # Create layout for shader editing
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(self.centralwidget)
        self.label.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignLeading
            | QtCore.Qt.AlignmentFlag.AlignLeft
            | QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.vertex_source_edit = QtWidgets.QTextEdit(self.centralwidget)
        self.vertex_source_edit.setAcceptRichText(False)
        self.vertex_source_edit.setObjectName("vertex_source_edit")
        self.verticalLayout.addWidget(self.vertex_source_edit)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_2 = QtWidgets.QLabel(self.centralwidget)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.fragSourceEdit = QtWidgets.QTextEdit(self.centralwidget)
        self.fragSourceEdit.setAcceptRichText(False)
        self.fragSourceEdit.setObjectName("frag_source_edit")
        self.verticalLayout_2.addWidget(self.fragSourceEdit)
        self.horizontalLayout.addLayout(self.verticalLayout_2)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.compileShaderBtn = QtWidgets.QPushButton(self.centralwidget)
        self.compileShaderBtn.clicked.connect(self.compileClick)
        self.compileShaderBtn.setObjectName("compile_shader_btn")
        self.verticalLayout_3.addWidget(self.compileShaderBtn)

        # Initialize the pyglet widget we'll draw to and lay out the window
        self.openGLWidget = PygletWidget(800, 400, self.centralwidget, self)
        self.openGLWidget.setMinimumSize(QtCore.QSize(800, 400))
        self.openGLWidget.setObjectName("openGLWidget")
        self.verticalLayout_3.addWidget(self.openGLWidget)
        self.gridLayout.addLayout(self.verticalLayout_3, 0, 0, 1, 1)

        MainWindow.setCentralWidget(self.centralwidget)

        # Set up the root top menu & status bar
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 820, 21))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        # Add menu bar menus, entries, and hotkeys
        self.actionOpen_Image = QtGui.QAction(MainWindow)
        self.actionOpen_Image.setObjectName("actionOpen_Image")
        self.actionOpen_Image.triggered.connect(self.loadImages)
        self.actionOpen_Image.setShortcut("Ctrl+I")

        self.actionOpen_Shader = QtGui.QAction(MainWindow)
        self.actionOpen_Shader.setObjectName("actionOpen_Shader")
        self.actionOpen_Shader.triggered.connect(self.loadShaders)
        self.actionOpen_Shader.setShortcut("Ctrl+O")

        self.actionSave_Shader = QtGui.QAction(MainWindow)
        self.actionSave_Shader.setObjectName("actionSave_Shader")
        self.actionSave_Shader.triggered.connect(self.saveShaders)
        self.actionSave_Shader.setStatusTip('Saves both Shader Files')
        self.actionSave_Shader.setShortcut("Ctrl+S")

        self.actionExit = QtGui.QAction(MainWindow)
        self.actionExit.triggered.connect(self.closeProgram)
        self.actionExit.setObjectName("actionExit")
        self.menuFile.addAction(self.actionOpen_Image)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionOpen_Shader)
        self.menuFile.addAction(self.actionSave_Shader)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionExit)
        self.menubar.addAction(self.menuFile.menuAction())

        self.imageMenu = QtWidgets.QMenu(self.menubar)
        self.imageMenu.setObjectName("imageMenu")
        self.noImageAction = QtGui.QAction(MainWindow)
        self.noImageAction.setDisabled(True)
        self.imageMenu.addAction(self.noImageAction)

        self.menubar.addAction(self.imageMenu.menuAction())

        # Perform localization & start accepting UI events
        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def compileClick(self) -> None:
        if not self.images:
            print("No images have been loaded.")
            return

        self.openGLWidget.makeCurrent()

        if self.sprite:
            self.sprite.delete()
            self.sprite = None

        if self.program:
            del self.program
            self.program = None

        try:
            vertex_ = pyglet.graphics.shader.Shader(self.vertex_source_edit.toPlainText(), 'vertex')
            fragment_ = pyglet.graphics.shader.Shader(self.fragSourceEdit.toPlainText(), 'fragment')

            self.program = pyglet.graphics.shader.ShaderProgram(vertex_, fragment_)

            if len(self.images) == 1:
                self.sprite = pyglet.sprite.Sprite(
                    self.images[0],
                    x=self.SPRITE_POSITION[0], y=self.SPRITE_POSITION[1],
                    group=self.group, batch=self.openGLWidget.batch,
                    program=self.program)
            else:
                textures = {image.shader_name: image.get_texture() for image in self.images}
                self.sprite = MultiTextureSprite(
                    textures,
                    x=self.SPRITE_POSITION[0], y=self.SPRITE_POSITION[1],
                    group=self.group, batch=self.openGLWidget.batch,
                    program=self.program)

            if self.program:
                print("Successfully compiled shader.")
        except pyglet.gl.lib.GLException as err:
            print(f"Failed to compile shader: {err}")
        except Exception as err:
            print("Unexpected error", err)
    def loadImages(self) -> None:
        options = self.get_file_dialog_options()
        fileNames, _ = QFileDialog.getOpenFileNames(
            self._window, "Select Image Files", "", "Image Files (*.png *.jpg *.jpeg *.bmp)",
            options=options)

        for fileName in fileNames:
            if not self.images:
                self.imageMenu.removeAction(self.noImageAction)

            action = QtGui.QAction(self._window)
            shader_name = f"sprite_texture{len(self.images)}"
            action.setText(f"{os.path.basename(fileName)} ({shader_name})")
            action.fileName = fileName
            action.setCheckable(True)
            action.setChecked(True)
            action.triggered.connect(lambda: self.removeImage(action))
            image = pyglet.image.load(fileName)
            image.shader_name = shader_name
            action.image = image
            self.images.append(image)
            self.imageMenu.addAction(action)

    def loadShaders(self) -> None:
        options = self.get_file_dialog_options()
        file_names, _ = QFileDialog.getOpenFileNames(
            self._window, "Load Shader Files", "", "Shader Files (*.vert *.frag *.txt)",
            options=options)

        for file_name in file_names:
            file_path = Path(file_name)
            ext = file_path.suffixes[-1]

            if ext == '.vert':
                dest = self.vertex_source_edit
            elif ext in ('.txt', '.frag'):
                dest = self.fragSourceEdit
            else:
                dest = self.fragSourceEdit

            text = file_path.read_text()
            dest.setText(text)

    def saveShaders(self) -> None:
        options = self.get_file_dialog_options()
        file_name, _ = QFileDialog.getSaveFileName(
            self._window, "Saving Both Shader Files (vert and frag)", "",
            options=options)

        if file_name:
            base_path = Path(file_name)
            vert_filename = base_path.with_suffix(".vert")
            frag_filename = base_path.with_suffix(".frag")

            vert_filename.write_text(self.vertex_source_edit.toPlainText())
            frag_filename.write_text(self.fragSourceEdit.toPlainText())

    def removeImage(self, actionWidget: QtGui.QAction) -> None:
        if self.imageMenu:
            self.imageMenu.removeAction(actionWidget)

            self.images.remove(actionWidget.image)

            # Re-order shader names
            for idx, image in enumerate(self.images):
                image.shader_name = f"sprite_texture{idx}"

            for action in self.imageMenu.actions():
                shader_name = action.image.shader_name
                action.setText(f"{os.path.basename(action.fileName)} ({shader_name})")

            if len(self.images) == 0:
                self.imageMenu.addAction(self.noImageAction)

    def closeProgram(self) -> None:
        app.exit()

    def retranslateUi(self, MainWindow: QtWidgets.QMainWindow) -> None:
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Sprite Shader Previewer"))
        self.label.setText(_translate("MainWindow", "Vertex Source:"))
        self.label_2.setText(_translate("MainWindow", "Fragment Source:"))
        self.compileShaderBtn.setText(_translate("MainWindow", "Compile Shaders"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.actionOpen_Image.setText(_translate("MainWindow", "Open Images"))
        self.actionOpen_Shader.setText(_translate("MainWindow", "Open Shader"))
        self.actionSave_Shader.setText(_translate("MainWindow", "Save Shader"))
        self.actionExit.setText(_translate("MainWindow", "Exit"))
        self.imageMenu.setTitle(_translate("MainWindow", "Images"))
        self.noImageAction.setText(_translate("MainWindow", "No Images Loaded"))

        self.vertex_source_edit.setText(default_vertex_src)
        self.fragSourceEdit.setText(default_frag_src)


class PygletWidget(QOpenGLWidget):
    _default_vertex_source = """#version 150 core
        in vec4 position;

        uniform WindowBlock
        {
            mat4 projection;
            mat4 view;
        } window;

        void main()
        {
            gl_Position = window.projection * window.view * position;
        }
    """
    _default_fragment_source = """#version 150 core
        out vec4 color;

        void main()
        {
            color = vec4(1.0, 0.0, 0.0, 1.0);
        }
    """

    def __init__(self, width, height, parent, mainWindow) -> None:
        super().__init__(parent)
        self.mainWindow = mainWindow
        self.setMinimumSize(width, height)

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self._pyglet_update)
        self.timer.setInterval(0)
        self.timer.start()

        self.zoom = 1.0

        self.elapsed = 0

        pyglet.clock.schedule_interval(self.update_time_uniform, 1 / 60.0)

    def wheelEvent(self, event: QWheelEvent) -> None:
        super().wheelEvent(event)
        if event.angleDelta().y() > 0:
            self.zoom *= 2
        else:
            self.zoom /= 2

        self.zoom = pyglet.math.clamp(self.zoom, 0.125, 40.0)
        self.view = pyglet.math.Mat4().scale(pyglet.math.Vec3(
            self.zoom, self.zoom, 1.0))

        event.accept()

    def update_time_uniform(self, dt: float) -> None:
        self.elapsed += dt

        if self.mainWindow.program:
            self.mainWindow.program.use()

            # Ignore time if it doesn't exist.
            try:
                self.mainWindow.program['time'] = self.elapsed
            except Exception:
                pass

            self.mainWindow.program.stop()

    def _pyglet_update(self) -> None:
        # Tick the pyglet clock, so scheduled events can work.
        pyglet.clock.tick()

        # Force widget to update, otherwise paintGL will not be called.
        self.update()  # self.updateGL() for pyqt5

    def paintGL(self) -> None:
        """Pyglet equivalent of on_draw event for window"""
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        self.batch.draw()

    def resizeGL(self, width: int, height: int) -> None:
        self.projection = pyglet.math.Mat4.orthogonal_projection(0, width, 0, height, -255, 255)

        self.viewport = 0, 0, width, height

    def initializeGL(self) -> None:
        """Call anything that needs a context to be created."""
        self._projection_matrix = pyglet.math.Mat4()
        self._view_matrix = pyglet.math.Mat4()

        self.batch = pyglet.graphics.Batch()

        self._default_program = pyglet.graphics.shader.ShaderProgram(
            pyglet.graphics.shader.Shader(self._default_vertex_source, 'vertex'),
            pyglet.graphics.shader.Shader(self._default_fragment_source, 'fragment'))

        self.ubo = self._default_program.uniform_blocks['WindowBlock'].create_ubo()

        self.view = pyglet.math.Mat4()
        self.projection = pyglet.math.Mat4.orthogonal_projection(0, self.width(), 0, self.height(), -255, 255)
        self.viewport = 0, 0, self.width(), self.height()

    @property
    def viewport(self) -> tuple[int, int, int, int]:
        """The Window viewport

        The Window viewport, expressed as (x, y, width, height).

        :return: The viewport size as a tuple of four ints.
        """
        return self._viewport

    @viewport.setter
    def viewport(self, values: tuple[int, int, int, int]) -> None:
        self._viewport = values
        pr = 1.0
        x, y, w, h = values
        pyglet.gl.glViewport(int(x * pr), int(y * pr), int(w * pr), int(h * pr))

    @property
    def projection(self) -> pyglet.math.Mat4:
        return self._projection_matrix

    @projection.setter
    def projection(self, matrix: pyglet.math.Mat4) -> None:
        with self.ubo as window_block:
            window_block.projection[:] = matrix

        self._projection_matrix = matrix

    @property
    def view(self) -> pyglet.math.Mat4:
        """The OpenGL window view matrix. Read-write.

        The default view is an identity matrix, but a custom
        :py:class:`pyglet.math.Mat4` instance can be set.
        Alternatively, you can supply a flat tuple of 16 values.
        """
        return self._view_matrix

    @view.setter
    def view(self, matrix: pyglet.math.Mat4) -> None:

        with self.ubo as window_block:
            window_block.view[:] = matrix

        self._view_matrix = matrix


def excepthook(exc_type, exc_value, exc_tb) -> None:
    """Replacement for Python's default exception handler function.

    See the following for more information:
    https://docs.python.org/3/library/sys.html#sys.excepthook
    """
    tb = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
    print(tb)


if __name__ == "__main__":
    # Create the base Qt application and initialize the UI
    app = QtWidgets.QApplication(sys.argv)
    ui = Ui_MainWindow(use_native_file_dialog=arguments.native_file_dialog)
    qt_window = QtWidgets.QMainWindow()
    ui.setupUi(qt_window)
    qt_window.show()

    # Replace the default exception handling *after* everything is
    # initialized to avoid swallowing fatal errors such as GL issues.
    sys.excepthook = excepthook

    # Start the application and return its exit code
    sys.exit(app.exec())



"""Application-wide functionality.

Applications
------------

Most applications need only call :func:`run` after creating one or more 
windows to begin processing events.  For example, a simple application 
consisting of one window is::

    import pyglet

    win = pyglet.window.Window()
    pyglet.app.run()


Events
======

To handle events on the main event loop, instantiate it manually.  The
following example exits the application as soon as any window is closed (the
default policy is to wait until all windows are closed)::

    event_loop = pyglet.app.EventLoop()

    @event_loop.event
    def on_window_close(window):
        event_loop.exit()

.. versionadded:: 1.1
"""
from __future__ import annotations

import sys
import weakref
import platform

import pyglet

from pyglet import compat_platform
from pyglet.app.base import EventLoop


_is_pyglet_doc_run = hasattr(sys, "is_pyglet_doc_run") and sys.is_pyglet_doc_run

if _is_pyglet_doc_run:
    from pyglet.app.base import PlatformEventLoop
else:
    if compat_platform == 'darwin':
        from pyglet.app.cocoa import CocoaPlatformEventLoop as PlatformEventLoop

        if platform.machine() == 'arm64' or pyglet.options["osx_alt_loop"]:
            from pyglet.app.cocoa import CocoaAlternateEventLoop as EventLoop
    elif compat_platform in ('win32', 'cygwin'):
        from pyglet.app.win32 import Win32EventLoop as PlatformEventLoop
    else:
        from pyglet.app.xlib import XlibEventLoop as PlatformEventLoop


class AppException(Exception):
    pass


windows = weakref.WeakSet()
"""Set of all open windows (including invisible windows).  Instances of
:class:`pyglet.window.Window` are automatically added to this set upon 
construction. The set uses weak references, so windows are removed from 
the set when they are no longer referenced or are closed explicitly.
"""


def run(interval: float | None = 1 / 60) -> None:
    """Begin processing events, scheduled functions and window updates.

    This is a convenience function, equivalent to::

        pyglet.app.event_loop.run(interval)

    """
    event_loop.run(interval)


def exit() -> None:
    """Exit the application event loop.

    Causes the application event loop to finish, if an event loop is currently
    running.  The application may not necessarily exit (for example, there may
    be additional code following the ``run`` invocation).

    This is a convenience function, equivalent to::

        pyglet.app.event_loop.exit()

    """
    event_loop.exit()


#: The global event loop. Applications can replace this
#: with their own subclass of :class:`EventLoop` before calling 
#: :meth:`EventLoop.run`.
event_loop = EventLoop()

#: The platform-dependent event loop. Applications are strongly discouraged
#: from subclassing or replacing this :class:`PlatformEventLoop` object.
platform_event_loop = PlatformEventLoop()



import pyglet
import pytest

_debug = False


class MockPlayer:
    def __init__(self, event_loop):
        self.event_loop = event_loop
        self.events = []
        self.pyclock = pyglet.clock.get_default()
        super().__init__()

    def dispatch_event(self, event_type, *args):
        super().dispatch_event(event_type, *args)
        if _debug:
            print('{}: event {} received @ {}'.format(
                self.__class__.__name__,
                event_type, self.pyclock.time()
            ))
        self.events.append((event_type, args))
        pyglet.clock.unschedule(self.event_loop.interrupt_event_loop)
        self.event_loop.interrupt_event_loop()

    def wait_for_event(self, timeout, *event_types):
        end_time = self.pyclock.time() + timeout

        while self.pyclock.time() < end_time:
            if _debug:
                print('{}: run for {} sec @ {}'.format(self.__class__.__name__,
                      end_time - self.pyclock.time(), self.pyclock.time()))
            self.event_loop.run_event_loop(duration=end_time - self.pyclock.time())
            if not self.events:
                continue
            event_type, args = self.events.pop()
            if event_type in event_types:
                return event_type, args
        return None, None

    def wait_for_all_events(self, timeout, *expected_events):
        now = self.pyclock.time()
        if _debug:
            print('Wait for events @ {}'.format(now))
        end_time = self.pyclock.time() + timeout
        expected_events = list(expected_events)
        received_events = []
        while expected_events:
            event_type, args = self.wait_for_event(timeout, *expected_events)
            if _debug:
                print('{}: got event {} @ {}'.format(self.__class__.__name__,
                      event_type, self.pyclock.time()))
            if event_type is None and self.pyclock.time() >= end_time:
                pytest.fail('Timeout before all events have been received. '
                            'Still waiting for: ' + ','.join(expected_events))
            elif event_type is not None:
                if event_type in expected_events:
                    expected_events.remove(event_type)
                received_events.append((event_type, args))
        return received_events

    def wait(self, timeout):
        now = self.pyclock.time()
        end_time = now + timeout
        while now - end_time < -0.005:
            duration = max(.01, end_time - now)
            self.event_loop.run_event_loop(duration=duration)
            now = self.pyclock.time()



import pyglet
import random

from pyglet.gl import GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_TRIANGLES
from pyglet.graphics.instance import InstanceSourceMixin

window = pyglet.window.Window(1920, 1080, vsync=False)

# Set example resource path.
pyglet.resource.path = ['../resources']
pyglet.resource.reindex()

# Load example image from resource path.
# Avoid using pyglet.image.load as it creates multiple textures.
# Resource will load images into a texture atlas, allowing significantly faster performance.
image = pyglet.resource.image("pyglet.png")

# Anchor point on an image is bottom left corner by default.
# Set to center point with anchor properties.
image.anchor_x = image.width // 2
image.anchor_y = image.height // 2

# Batching allows rendering groups of objects all at once instead of drawing one by one.
batch = pyglet.graphics.Batch()

scales = [1.0, 0.75, 0.5, 0.25]


class SpriteInstanceSource(pyglet.sprite.Sprite, InstanceSourceMixin):
    """Subclass example."""

    def __init__(self, img, attributes, x=0, y=0, z=0, blend_src=GL_SRC_ALPHA, blend_dest=GL_ONE_MINUS_SRC_ALPHA, batch=None,
                 group=None, subpixel=False):
        self.instance_attributes = attributes
        InstanceSourceMixin.__init__(self, attributes)
        pyglet.sprite.Sprite.__init__(self, img, x, y, z, blend_src, blend_dest, batch, group, subpixel)


    def _create_vertex_list(self):
        self._vertex_list = self.program.vertex_list_instanced_indexed(
            4, GL_TRIANGLES, [0, 1, 2, 0, 2, 3], self.instance_attributes,
            self._batch, self._group,
            position=('f', self._get_vertices()),
            colors=('Bn', (*self._rgb, int(self._opacity)) * (1 if 'colors' in self.instance_attributes else 4)),
            translate=('f', (self._x, self._y, self._z) * (1 if 'translate' in self.instance_attributes else 4)),
            scale=('f', (self._scale*self._scale_x, self._scale*self._scale_y) * (1 if 'scale' in self.instance_attributes else 4)),
            rotation=('f', (self._rotation,) * (1 if 'rotation' in self.instance_attributes else 4)),
            tex_coords=('f', self._texture.tex_coords))


sprites = []
# Create 1000 sprites at various scales.
sprite = SpriteInstanceSource(image, ('translate', 'colors'),
                              x=random.randint(0, window.width),
                              y=random.randint(0, window.height),
                              batch=batch)

instance = sprite.create_instance(translate=(100, 50, 0), colors=(255, 255, 255, 255))

for i in range(10):
    sprite_instance = sprite.create_instance(
        translate=(random.randint(0, window.width), random.randint(0, window.height), 0),
        colors=(random.randint(0, 255), random.randint(0, 255), random.randint(0, 255), 255),
    )

    #Randomize scale.
    #sprite_instance.scale = random.choice(scales)

    #Random color multiplier.
    #sprite_instance.colors = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))

    # Add sprites to keep in memory, like a list. Otherwise they will get GC'd when out of scope.
    sprites.append(sprite_instance)

#fps = pyglet.window.FPSDisplay(window)

@window.event
def on_key_press(symbol, modifiers):
    if symbol == pyglet.window.key.SPACE:
        instance.translate = (random.randint(0, window.width), random.randint(0, window.height), 0)
        # instance.translate[:] = (random.randint(0, window.width), random.randint(0, window.height), 0)
    elif symbol == pyglet.window.key.DELETE:
        del_instance = instance_sprite.pop()
        sprites.remove(del_instance)
    elif symbol == pyglet.window.key.INSERT:
        sprite_instance = instance_sprite.create(
            translate=(random.randint(0, window.width), random.randint(0, window.height), 0),
            colors=(random.randint(0, 255), random.randint(0, 255), random.randint(0, 255), 255),
        )
        sprites.append(sprite_instance)

@window.event
def on_draw():
    window.clear()
    batch.draw()
    #fps.draw()


pyglet.app.run(0)




from __future__ import annotations

import os

import pyglet

from pyglet.gl import GL_TRIANGLES
from pyglet.util import asstr

from . import ModelDecodeException, ModelDecoder
from .base import SimpleMaterial, Mesh, Primitive, Attribute, Node, Scene
from .. import Model, MaterialGroup, TexturedMaterialGroup
from ...graphics import Batch, Group


def _new_mesh(name, material):
    # The three primitive types used in .obj files:
    attributes = [Attribute('POSITION', 'f', 'VEC3', 0, []),
                  Attribute('NORMAL', 'f', 'VEC3', 0, []),
                  Attribute('TEXCOORD_0', 'f', 'VEC3', 0, [])]
    primitive = Primitive(attributes=attributes, indices=None, material=material, mode=GL_TRIANGLES)
    mesh = Mesh(primitives=[primitive], name=name)
    return mesh


def load_material_library(filename):
    file = open(filename, 'r')

    name = None
    diffuse = [1.0, 1.0, 1.0]
    ambient = [1.0, 1.0, 1.0]
    specular = [1.0, 1.0, 1.0]
    emission = [0.0, 0.0, 0.0]
    shininess = 100.0
    opacity = 1.0
    texture_name = None

    matlib = {}

    for line in file:
        if line.startswith('#'):
            continue
        values = line.split()
        if not values:
            continue

        if values[0] == 'newmtl':
            if name is not None:
                # save previous material
                for item in (diffuse, ambient, specular, emission):
                    item.append(opacity)
                matlib[name] = SimpleMaterial(name, diffuse, ambient, specular, emission, shininess, texture_name)
            name = values[1]

        elif name is None:
            raise ModelDecodeException(f'Expected "newmtl" in {filename}')

        try:
            if values[0] == 'Kd':
                diffuse = list(map(float, values[1:]))
            elif values[0] == 'Ka':
                ambient = list(map(float, values[1:]))
            elif values[0] == 'Ks':
                specular = list(map(float, values[1:]))
            elif values[0] == 'Ke':
                emission = list(map(float, values[1:]))
            elif values[0] == 'Ns':
                shininess = float(values[1])            # Blender exports 1~1000
                shininess = (shininess * 128) / 1000    # Normalize to 1~128 for OpenGL
            elif values[0] == 'd':
                opacity = float(values[1])
            elif values[0] == 'map_Kd':
                texture_name = values[1]

        except BaseException as ex:
            raise ModelDecodeException('Parsing error in {0}.'.format((filename, ex)))

    file.close()

    for item in (diffuse, ambient, specular, emission):
        item.append(opacity)

    matlib[name] = SimpleMaterial(name, diffuse, ambient, specular, emission, shininess, texture_name)

    return matlib


def parse_obj_file(filename, file=None) -> list[Mesh]:
    materials = {}
    meshes = []

    location = os.path.dirname(filename)

    try:
        if file is None:
            with open(filename, 'r') as f:
                file_contents = f.read()
        else:
            file_contents = asstr(file.read())
    except (UnicodeDecodeError, OSError):
        raise ModelDecodeException

    material = None
    mesh = None

    normals = [[0., 0., 0.]]
    tex_coords = [[0., 0.]]
    vertices = [[0., 0., 0.]]

    diffuse = [1.0, 1.0, 1.0, 1.0]
    ambient = [1.0, 1.0, 1.0, 1.0]
    specular = [1.0, 1.0, 1.0, 1.0]
    emission = [0.0, 0.0, 0.0, 1.0]
    shininess = 100.0

    default_material = SimpleMaterial("Default", diffuse, ambient, specular, emission, shininess)

    for line in file_contents.splitlines():

        if line.startswith('#'):
            continue
        values = line.split()
        if not values:
            continue

        if values[0] == 'v':
            vertices.append(list(map(float, values[1:4])))
        elif values[0] == 'vn':
            normals.append(list(map(float, values[1:4])))
        elif values[0] == 'vt':
            tex_coords.append(list(map(float, values[1:3])))

        elif values[0] == 'mtllib':
            material_abspath = os.path.join(location, values[1])
            materials = load_material_library(filename=material_abspath)            

        elif values[0] in ('usemtl', 'usemat'):
            material = materials.get(values[1])
            if mesh is not None:
                mesh.primitives[0].material = material

        elif values[0] == 'o':
            mesh = _new_mesh(name=values[1], material=default_material)
            meshes.append(mesh)

        elif values[0] == 'f':
            if material is None:
                material = SimpleMaterial()
            if mesh is None:
                mesh = _new_mesh(name='unknown', material=material)
                meshes.append(mesh)

            # For fan triangulation, remember first and latest vertices
            n1 = None
            nlast = None
            t1 = None
            tlast = None
            v1 = None
            vlast = None

            for i, v in enumerate(values[1:]):
                v_i, t_i, n_i = (list(map(int, [j or 0 for j in v.split('/')])) + [0, 0])[:3]
                if v_i < 0:
                    v_i += len(vertices) - 1
                if t_i < 0:
                    t_i += len(tex_coords) - 1
                if n_i < 0:
                    n_i += len(normals) - 1

                mesh.primitives[0].attributes[0].array += vertices[v_i]
                mesh.primitives[0].attributes[1].array += normals[n_i]
                mesh.primitives[0].attributes[2].array += tex_coords[t_i]

                if i >= 3:
                    mesh.primitives[0].attributes[0].array += v1 + vlast
                    mesh.primitives[0].attributes[1].array += n1 + nlast
                    mesh.primitives[0].attributes[2].array += t1 + tlast

                if i == 0:
                    n1 = normals[n_i]
                    t1 = tex_coords[t_i]
                    v1 = vertices[v_i]
                nlast = normals[n_i]
                tlast = tex_coords[t_i]
                vlast = vertices[v_i]

        for mesh in meshes:
            for primitive in mesh.primitives:
                for attribute in primitive.attributes:
                    attribute.count = len(attribute.array) // 3

    return meshes


class OBJScene(Scene):

    def create_models(self, batch: Batch, group: Group | None = None) -> list[Model]:
        vertex_lists = []
        groups = []
        for node in self.nodes:
            for mesh in node.meshes:
                material = mesh.primitives[0].material
                count = mesh.primitives[0].attributes[0].count

                if material.texture_name:
                    program = pyglet.model.get_default_textured_shader()
                    texture = pyglet.resource.texture(material.texture_name)
                    matgroup = TexturedMaterialGroup(material, program, texture, parent=group)
                else:
                    program = pyglet.model.get_default_shader()
                    matgroup = MaterialGroup(material, program, parent=group)

                data = {a.name: (a.fmt, a.array) for a in mesh.primitives[0].attributes}
                # Add additional material data:
                data['COLOR_0'] = 'f', material.diffuse * count

                vertex_lists.append(program.vertex_list(count, GL_TRIANGLES, batch, matgroup, **data))
                groups.append(matgroup)

        return [Model(vertex_lists=vertex_lists, groups=groups, batch=batch)]


###################################################
#   Decoder definitions start here:
###################################################

class OBJModelDecoder(ModelDecoder):
    def get_file_extensions(self):
        return ['.obj']

    def decode(self, filename, file):

        mesh_list = parse_obj_file(filename=filename, file=file)
        return OBJScene(nodes=[Node(meshes=mesh_list)])


def get_decoders():
    return [OBJModelDecoder()]


def get_encoders():
    return []



"""Tests for window settings."""
import pytest
import time

from tests.base.interactive import InteractiveTestCase
from tests.interactive.window import window_util

from pyglet.gl import *
from pyglet import image
from pyglet.window import key, Window, ImageMouseCursor
from pyglet.window.event import WindowEventLogger


@pytest.mark.requires_user_action
class WINDOW_SET_EXCLUSIVE_KEYBOARD(InteractiveTestCase):
    """Test that exclusive keyboard mode can be set.

    Expected behaviour:
        One window will be opened.  Press 'e' to enable exclusive mode and 'E'
        to disable exclusive mode.

        In exclusive mode:
        - Pressing system keys, the Expose keys, etc., should have no effect
        besides displaying as keyboard events.
            - On OS X, the Power switch is not disabled (though this is possible
            if desired, see source).
            - On OS X, the menu bar and dock will disappear during keyboard
            exclusive mode.
            - On Windows, only Alt+Tab is disabled.  A user can still switch away
            using Ctrl+Escape, Alt+Escape, the Windows key or Ctrl+Alt+Del.
        - Switching to another application (i.e., with the mouse) should make
        these keys work normally again until this application regains focus.

        Close the window or press ESC to end the test.
    """
    def on_key_press(self, symbol, modifiers):
        print('Pressed %s with modifiers %s' % \
            (key.symbol_string(symbol), key.modifiers_string(modifiers)))

        if symbol == key.E:
            exclusive = not (modifiers & key.MOD_SHIFT)
            self.w.set_exclusive_keyboard(exclusive)
            print('Exclusive keyboard is now %r' % exclusive)

    def on_key_release(self, symbol, modifiers):
        print('Released %s with modifiers %s' % \
            (key.symbol_string(symbol), key.modifiers_string(modifiers)))

    def test_set_exclusive_keyboard(self):
        self.width, self.height = 200, 200
        self.w = w = Window(self.width, self.height)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class WINDOW_SET_EXCLUSIVE_MOUSE(InteractiveTestCase):
    """Test that exclusive mouse mode can be set.

    Expected behaviour:
        One window will be opened.  Press 'e' to enable exclusive mode and 'E'
        to disable exclusive mode.

        In exclusive mode:
        - the mouse cursor should be invisible
        - moving the mouse should generate events with bogus x,y but correct
        dx and dy.
        - it should not be possible to switch applications with the mouse
        - if application loses focus (i.e., with keyboard), the mouse should
        operate normally again until focus is returned to the app, in which
        case it should hide again.

        Close the window or press ESC to end the test.
    """
    def on_key_press(self, symbol, modifiers):
        if symbol == key.E:
            exclusive = not (modifiers & key.MOD_SHIFT)
            self.w.set_exclusive_mouse(exclusive)
            print('Exclusive mouse is now %r' % exclusive)

    def on_mouse_motion(self, x, y, dx, dy):
        print('on_mousemotion(x=%f, y=%f, dx=%f, dy=%f)' % (x, y, dx, dy))

    def test_set_exclusive_mouse(self):
        self.width, self.height = 200, 200
        self.w = w = Window(self.width, self.height)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class WINDOW_SET_FULLSCREEN(InteractiveTestCase):
    """Test that window can be set fullscreen and back again.

    Expected behaviour:
        One window will be opened.  

        - press "f" to enter fullscreen mode.
        - press "g" to leave fullscreen mode.

        All events will be printed to the terminal.

        Close the window or press ESC to end the test.
    """
    def on_key_press(self, symbol, modifiers):
        if symbol == key.F:
            print('Setting fullscreen.')
            self.w.set_fullscreen(True)
        elif symbol == key.G:
            print('Leaving fullscreen.')
            self.w.set_fullscreen(False)

    def on_expose(self):
        glClearColor(1, 0, 0, 1)
        glClear(GL_COLOR_BUFFER_BIT)
        self.w.flip()

    def test_set_fullscreen(self):
        self.w = w = Window(200, 200)
        try:
            w.push_handlers(self)
            w.push_handlers(WindowEventLogger())
            self.on_expose()
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_validation
class WINDOW_SET_ICON(InteractiveTestCase):
    """Test that window icon can be set.

    Expected behaviour:
        One window will be opened.  It will have an icon depicting a yellow
        "A".
    """
    def test_set_icon(self):
        self.width, self.height = 200, 200
        self.w = w = Window(self.width, self.height)
        try:
            w.set_icon(image.load(self.get_test_data_file('images', 'icon1.png')))
            w.dispatch_events()
            self.user_verify('Does the window have a yellow A icon?', take_screenshot=False)
        finally:
            w.close()


@pytest.mark.requires_user_validation
class WINDOW_SET_ICON_SIZES(InteractiveTestCase):
    """Test that window icon can be set for multiple sizes.

    Expected behaviour:
        One window will be opened.  The window's icon depends on the icon
        size:

        16x16 icon is a yellow "1"
        32x32 icon is a purple "2"
        48x48 icon is a cyan "3"
        72x72 icon is a red "4"
        128x128 icon is a blue "5"

        For other sizes, the operating system may select the closest match and
        scale it (Linux, Windows), or interpolate between two or more images
        (Mac OS X).
    """
    def test_set_icon_sizes(self):
        self.width, self.height = 200, 200
        self.w = w = Window(self.width, self.height)
        try:
            w.set_icon(image.load(self.get_test_data_file('images', 'icon_size1.png')),
                    image.load(self.get_test_data_file('images', 'icon_size2.png')),
                    image.load(self.get_test_data_file('images', 'icon_size3.png')),
                    image.load(self.get_test_data_file('images', 'icon_size4.png')),
                    image.load(self.get_test_data_file('images', 'icon_size5.png')))
            w.dispatch_events()
            self.user_verify('Does the window have the icon corresponding to the correct size?', take_screenshot=False)
        finally:
            w.close()


@pytest.mark.requires_user_action
class WINDOW_SET_LOCATION(InteractiveTestCase):
    """Test that window location can be set.

    Expected behaviour:
        One window will be opened.  The window's location will be printed
        to the terminal.  

        - Use the arrow keys to move the window.

        Close the window or press ESC to end the test.
    """
    def on_key_press(self, symbol, modifiers):
        x, y = self.w.get_location()
        if symbol == key.LEFT:
            x -= 10
        if symbol == key.RIGHT:
            x += 10
        if symbol == key.UP:
            y -= 10
        if symbol == key.DOWN:
            y += 10
        self.w.set_location(x, y)
        print('Window location set to %dx%d.' % (x, y))
        print('Window location now: %dx%d.' % self.w.get_location())
        self.assertSequenceEqual((x, y), self.w.get_location())

    def test_set_location(self):
        self.w = w = Window(200, 200)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class WINDOW_SET_MIN_MAX_SIZE(InteractiveTestCase):
    """Test that minimum and maximum window size can be set.

    Expected behaviour:
        One window will be opened.  The window's dimensions will be printed
        to the terminal.  Initially the window has no minimum or maximum
        size (besides any OS-enforced limit).

        - press "n" to set the minimum size to be the current size.
        - press "x" to set the maximum size to be the current size.

        You should see a green border inside the window but no red.

        Close the window or press ESC to end the test.
    """
    def on_resize(self, width, height):
        print('Window size is %dx%d.' % (width, height))
        self.width, self.height = width, height

    def on_key_press(self, symbol, modifiers):
        if symbol == key.N:
            self.w.set_minimum_size(self.width, self.height)
            print('Minimum size set to %dx%d.' % (self.width, self.height))
        elif symbol == key.X:
            self.w.set_maximum_size(self.width, self.height)
            print('Maximum size set to %dx%d.' % (self.width, self.height))

    def test_min_max_size(self):
        self.width, self.height = 200, 200
        self.w = w = Window(self.width, self.height, resizable=True)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                window_util.draw_client_border(w)
                w.flip()
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class WINDOW_SET_MOUSE_CURSOR(InteractiveTestCase):
    """Test that image mouse cursor can be set.

    Expected behaviour:
        One window will be opened.  The mouse cursor in the window will be
        a custom cursor.

        Close the window or press ESC to end the test.
    """
    def on_mouse_motion(self, x, y, dx, dy):
        print('on_mousemotion(x=%f, y=%f, dx=%f, dy=%f)' % (x, y, dx, dy))

    def test_set_mouse_cursor(self):
        self.width, self.height = 200, 200
        self.w = w = Window(self.width, self.height)
        try:
            img = image.load(self.get_test_data_file('images', 'cursor.png'))
            w.set_mouse_cursor(ImageMouseCursor(img, 4, 28))
            w.push_handlers(self)
            glClearColor(1, 1, 1, 1)
            while not w.has_exit:
                glClear(GL_COLOR_BUFFER_BIT)
                w.flip()
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class WINDOW_SET_MOUSE_PLATFORM_CURSOR(InteractiveTestCase):
    """Test that mouse cursor can be set to a platform-dependent image.

    Expected behaviour:
        One window will be opened.  Press the left and right arrow keys to cycle
        through the system mouse cursors.  The current cursor selected will
        be printed to the terminal.

        Note that not all cursors are unique on each platform; for example,
        if a platform doesn't define a cursor for a given name, a suitable
        replacement (e.g., a plain arrow) will be used instead.

        Close the window or press ESC to end the test.
    """
    i = 0
    def on_key_press(self, symbol, modifiers):
        names = [
            self.w.CURSOR_DEFAULT,
            self.w.CURSOR_CROSSHAIR,
            self.w.CURSOR_HAND,
            self.w.CURSOR_HELP,
            self.w.CURSOR_NO,
            self.w.CURSOR_SIZE,
            self.w.CURSOR_SIZE_UP,
            self.w.CURSOR_SIZE_UP_RIGHT,
            self.w.CURSOR_SIZE_RIGHT,
            self.w.CURSOR_SIZE_DOWN_RIGHT,
            self.w.CURSOR_SIZE_DOWN,
            self.w.CURSOR_SIZE_DOWN_LEFT,
            self.w.CURSOR_SIZE_LEFT,
            self.w.CURSOR_SIZE_UP_LEFT,
            self.w.CURSOR_SIZE_UP_DOWN,
            self.w.CURSOR_SIZE_LEFT_RIGHT,
            self.w.CURSOR_TEXT,
            self.w.CURSOR_WAIT,
            self.w.CURSOR_WAIT_ARROW,
        ]
        if symbol == key.ESCAPE:
            self.w.on_close()
        if symbol == key.RIGHT:
            self.i = (self.i + 1) % len(names)
        elif symbol == key.LEFT:
            self.i = (self.i - 1) % len(names)
        cursor = self.w.get_system_mouse_cursor(names[self.i])
        self.w.set_mouse_cursor(cursor)
        print('Set cursor to "%s"' % names[self.i])

        return True

    def test_set_visible(self):
        self.width, self.height = 200, 200
        self.w = w = Window(self.width, self.height)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                glClear(GL_COLOR_BUFFER_BIT)
                w.flip()
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class WINDOW_SET_MOUSE_VISIBLE(InteractiveTestCase):
    """Test that mouse cursor can be made visible and hidden.

    Expected behaviour:
        One window will be opened.  Press 'v' to hide mouse cursor and 'V' to
        show mouse cursor.  It should only affect the mouse when within the
        client area of the window.

        Close the window or press ESC to end the test.
    """
    def on_key_press(self, symbol, modifiers):
        if symbol == key.V:
            visible = (modifiers & key.MOD_SHIFT)
            self.w.set_mouse_visible(visible)
            print('Mouse is now %s' % (visible and 'visible' or 'hidden'))

    def on_mouse_motion(self, x, y, dx, dy):
        print('on_mousemotion(x=%f, y=%f, dx=%f, dy=%f)' % (x, y, dx, dy))

    def test_set_visible(self):
        self.width, self.height = 200, 200
        self.w = w = Window(self.width, self.height)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class WINDOW_SET_SIZE(InteractiveTestCase):
    """Test that window size can be set.

    Expected behaviour:
        One window will be opened.  The window's dimensions will be printed
        to the terminal.

        - press "x" to increase the width
        - press "X" to decrease the width
        - press "y" to increase the height
        - press "Y" to decrease the height

        You should see a green border inside the window but no red.

        Close the window or press ESC to end the test.
    """
    def on_key_press(self, symbol, modifiers):
        delta = 20
        if modifiers & key.MOD_SHIFT:
            delta = -delta
        if symbol == key.X:
            self.width += delta
        elif symbol == key.Y:
            self.height += delta
        self.w.set_size(self.width, self.height)
        print('Window size set to %dx%d.' % (self.width, self.height))

    def test_set_size(self):
        self.width, self.height = 200, 200
        self.w = w = Window(self.width, self.height, resizable=True)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                window_util.draw_client_border(w)
                w.flip()
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_validation
class WINDOW_SET_VISIBLE(InteractiveTestCase):
    """Test that the window can be hidden and shown.

    Expected behaviour:
        One window will be opened. It will toggle between
        hidden and shown.
    """
    def test_set_visible(self):
        w = Window(200, 200)
        try:
            w.push_handlers(WindowEventLogger())
            w.dispatch_events()
            self.user_verify('Is the window visible?', take_screenshot=False)

            w.set_visible(False)
            w.dispatch_events()
            self.user_verify('Is the window no longer visible?', take_screenshot=False)

            w.set_visible(True)
            w.dispatch_events()
            self.user_verify('Is the window visible again?', take_screenshot=False)

        finally:
            w.close()


@pytest.mark.requires_user_action
class WINDOW_SET_VSYNC(InteractiveTestCase):
    """Test that vsync can be set.

    Expected behaviour:
        A window will alternate between red and green fill.

        - Press "v" to toggle vsync on/off.  "Tearing" should only be visible
            when vsync is off (as indicated at the terminal).

        Not all video drivers support vsync.  On Linux, check the output of
        `tools/info.py`:

        - If GLX_SGI_video_sync extension is present, should work as expected.
        - If GLX_MESA_swap_control extension is present, should work as expected.
        - If GLX_SGI_swap_control extension is present, vsync can be enabled,
            but once enabled, it cannot be switched off (there will be no error
            message).
        - If none of these extensions are present, vsync is not supported by
            your driver, but no error message or warning will be printed.

        Close the window or press ESC to end the test.
    """
    colors = [(1, 0, 0, 1), (0, 1, 0, 1)]
    color_index = 0

    def open_window(self):
        return Window(200, 200, vsync=False)

    def on_key_press(self, symbol, modifiers):
        if symbol == key.V:
            vsync = not self.w1.vsync
            self.w1.set_vsync(vsync)
            print('vsync is %r' % self.w1.vsync)

    def draw_window(self, window, colour):
        window.switch_to()
        glClearColor(*colour)
        glClear(GL_COLOR_BUFFER_BIT)
        window.flip()

    def test_open_window(self):
        self.w1 = self.open_window()
        try:
            self.w1.push_handlers(self)
            print('vsync is %r' % self.w1.vsync)
            while not self.w1.has_exit:
                self.color_index = 1 - self.color_index
                self.draw_window(self.w1, self.colors[self.color_index])
                self.w1.dispatch_events()
        finally:
            self.w1.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class WINDOW_SET_CAPTION(InteractiveTestCase):
    """Test that the window caption can be set.

    Expected behaviour:
        Two windows will be opened, one with the caption "Window caption 1"
        counting up every second; the other with a Unicode string including
        some non-ASCII characters.

        Press escape or close either window to finished the test.
    """
    def test_caption(self):
        try:
            w1 = Window(400, 200, resizable=True)
            w2 = Window(400, 200, resizable=True)
            count = 1
            w1.set_caption('Window caption %d' % count)
            w2.set_caption(u'\u00bfHabla espa\u00f1ol?')
            last_time = time.time()
            while not (w1.has_exit or w2.has_exit):
                if time.time() - last_time > 1:
                    count += 1
                    w1.set_caption('Window caption %d' % count)
                    last_time = time.time()
                w1.dispatch_events()
                w2.dispatch_events()
        finally:
            w1.close()
            w2.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class WINDOW_FIXED_SET_SIZE(InteractiveTestCase):
    """Test that a non-resizable window's size can be set.

    Expected behaviour:
        One window will be opened.  The window's dimensions will be printed
        to the terminal.

        - press "x" to increase the width
        - press "X" to decrease the width
        - press "y" to increase the height
        - press "Y" to decrease the height

        You should see a green border inside the window but no red.

        Close the window or press ESC to end the test.
    """
    def on_key_press(self, symbol, modifiers):
        delta = 20
        if modifiers & key.MOD_SHIFT:
            delta = -delta
        if symbol == key.X:
            self.width += delta
        elif symbol == key.Y:
            self.height += delta
        self.w.set_size(self.width, self.height)
        print('Window size set to %dx%d.' % (self.width, self.height))

    def test_set_size(self):
        self.width, self.height = 200, 200
        self.w = w = Window(self.width, self.height)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                window_util.draw_client_border(w)
                w.flip()
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)





from __future__ import annotations

from pyglet.libs.darwin import cocoapy


# This class is a wrapper around NSCursor which prevents us from
# sending too many hide or unhide messages in a row.  Apparently
# NSCursor treats them like retain/release messages, which can be
# problematic when we are e.g. switching between window & fullscreen.
class SystemCursor:
    cursor_is_hidden = False

    @classmethod
    def hide(cls) -> None:  # noqa: ANN102
        if not cls.cursor_is_hidden:
            cocoapy.send_message('NSCursor', 'hide')
            cls.cursor_is_hidden = True

    @classmethod
    def unhide(cls) -> None:  # noqa: ANN102
        if cls.cursor_is_hidden:
            cocoapy.send_message('NSCursor', 'unhide')
            cls.cursor_is_hidden = False



"""OpenGL Buffer Objects.

:py:class:`~BufferObject` and a :py:class:`~BackedBufferObject` are provided.
The first is a lightweight abstraction over an OpenGL buffer, as created
with ``glGenBuffers``. The backed buffer object is similar, but provides a
full mirror of the data in CPU memory. This allows for delayed uploading of
changes to GPU memory, which can improve performance is some cases.
"""
from __future__ import annotations

import abc
import ctypes
import sys
from functools import lru_cache
from typing import TYPE_CHECKING, Sequence, Type

from _ctypes import Array, _Pointer, _SimpleCData

import pyglet
from pyglet.gl.gl import (
    GL_ARRAY_BUFFER,
    GL_DYNAMIC_DRAW,
    GL_ELEMENT_ARRAY_BUFFER,
    GL_MAP_READ_BIT,
    GL_MAP_WRITE_BIT,
    GL_MAP_COHERENT_BIT,
    GL_MAP_PERSISTENT_BIT,
    GL_WRITE_ONLY,
    GLubyte,
    GLuint,
    glBindBuffer,
    glBufferData,
    glBufferStorage,
    glBufferSubData,
    glDeleteBuffers,
    glGenBuffers,
    glMapBuffer,
    glMapBufferRange,
    glUnmapBuffer,
)

if TYPE_CHECKING:
    from pyglet.gl import Context
    from pyglet.graphics.shader import Attribute

CTypesDataType = Type[_SimpleCData]
CTypesPointer = _Pointer


class AbstractBuffer:
    """Abstract buffer of byte data.

    Attributes:
        size:
            Size of buffer, in bytes
        ptr:
            Memory offset of the buffer, as used by the ``glVertexPointer`` family of functions
    """

    ptr: int = 0
    size: int = 0

    @abc.abstractmethod
    def bind(self, target: int = GL_ARRAY_BUFFER) -> None:
        """Bind this buffer to an OpenGL target."""

    @abc.abstractmethod
    def unbind(self) -> None:
        """Reset the buffer's OpenGL target."""

    @abc.abstractmethod
    def set_data(self, data: Sequence[int] | CTypesPointer) -> None:
        """Set the entire contents of the buffer.

        Args:
            data:
                The byte array to set.

        """

    @abc.abstractmethod
    def set_data_region(self, data: Sequence[int] | CTypesPointer, start: int, length: int) -> None:
        """Set part of the buffer contents.

        Args:
            data:
                The byte array of data to set
            start:
                Offset to start replacing data
            length:
                Length of region to replace

        """

    @abc.abstractmethod
    def map(self) -> CTypesPointer[ctypes.c_ubyte]:
        """Map the entire buffer into system memory.

        The mapped region must be subsequently unmapped with `unmap` before
        performing any other operations on the buffer.

        Returns:
            Pointer to the mapped block in memory
        """

    @abc.abstractmethod
    def unmap(self) -> None:
        """Unmap a previously mapped memory block."""

    def resize(self, size: int) -> None:
        """Resize the buffer to a new size.

        Args:
            size:
                New size of the buffer, in bytes

        """

    @abc.abstractmethod
    def delete(self) -> None:
        """Delete this buffer, reducing system resource usage."""


class BufferObject(AbstractBuffer):
    """Lightweight representation of an OpenGL Buffer Object.

    The data in the buffer is not replicated in any system memory (unless it
    is done so by the video driver).  While this can reduce system memory usage,
    performing multiple small updates to the buffer can be relatively slow.
    The target of the buffer is ``GL_ARRAY_BUFFER`` internally to avoid
    accidentally overriding other states when altering the buffer contents.
    The intended target can be set when binding the buffer.
    """

    id: int
    size: int
    usage: int
    _context: Context | None

    def __init__(self, size: int, usage: int = GL_DYNAMIC_DRAW) -> None:
        """Initialize the BufferObject with the given size and draw usage.

        Buffer data is cleared on creation.
        """
        self.size = size
        self.usage = usage
        self._context = pyglet.gl.current_context

        buffer_id = GLuint()
        glGenBuffers(1, buffer_id)
        self.id = buffer_id.value

        glBindBuffer(GL_ARRAY_BUFFER, self.id)
        data = (GLubyte * self.size)()
        glBufferData(GL_ARRAY_BUFFER, self.size, data, self.usage)

    def invalidate(self) -> None:
        glBufferData(GL_ARRAY_BUFFER, self.size, None, self.usage)

    def bind(self, target: int = GL_ARRAY_BUFFER) -> None:
        glBindBuffer(target, self.id)

    def unbind(self) -> None:
        glBindBuffer(GL_ARRAY_BUFFER, 0)

    def bind_to_index_buffer(self) -> None:
        """Binds this buffer as an index buffer on the active vertex array."""
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.id)

    def set_data(self, data: Sequence[int] | CTypesPointer) -> None:
        glBindBuffer(GL_ARRAY_BUFFER, self.id)
        glBufferData(GL_ARRAY_BUFFER, self.size, data, self.usage)

    def set_data_region(self, data: Sequence[int] | CTypesPointer, start: int, length: int) -> None:
        glBindBuffer(GL_ARRAY_BUFFER, self.id)
        glBufferSubData(GL_ARRAY_BUFFER, start, length, data)

    def map(self) -> CTypesPointer[ctypes.c_byte]:
        glBindBuffer(GL_ARRAY_BUFFER, self.id)
        return ctypes.cast(glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY),
                           ctypes.POINTER(ctypes.c_byte * self.size)).contents

    def map_range(self, start: int, size: int, ptr_type: type[CTypesPointer]) -> CTypesPointer:
        glBindBuffer(GL_ARRAY_BUFFER, self.id)
        return ctypes.cast(glMapBufferRange(GL_ARRAY_BUFFER, start, size, GL_MAP_WRITE_BIT), ptr_type).contents

    def unmap(self) -> None:
        glUnmapBuffer(GL_ARRAY_BUFFER)

    def delete(self) -> None:
        glDeleteBuffers(1, GLuint(self.id))
        self.id = None

    def __del__(self) -> None:
        if self.id is not None:
            try:
                self._context.delete_buffer(self.id)
                self.id = None
            except (AttributeError, ImportError):
                pass  # Interpreter is shutting down

    def resize(self, size: int) -> None:
        # Map, create a copy, then reinitialize.
        temp = (ctypes.c_byte * size)()

        glBindBuffer(GL_ARRAY_BUFFER, self.id)
        data = glMapBufferRange(GL_ARRAY_BUFFER, 0, self.size, GL_MAP_READ_BIT)
        ctypes.memmove(temp, data, min(size, self.size))
        glUnmapBuffer(GL_ARRAY_BUFFER)

        self.size = size
        glBufferData(GL_ARRAY_BUFFER, self.size, temp, self.usage)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, size={self.size})"


class BackedBufferObject(BufferObject):
    """A buffer with system-memory backed store.

    Updates to the data via ``set_data`` and ``set_data_region`` will be held
    in system memory until ``commit`` is called.  The advantage is that fewer
    OpenGL calls are needed, which can increase performance at the expense of
    system memory.
    """
    data: CTypesDataType
    data_ptr: int
    _dirty_min: int
    _dirty_max: int
    _dirty: bool
    stride: int
    count: int
    ctype: CTypesDataType

    def __init__(self, size: int, c_type: CTypesDataType, stride: int, count: int,  # noqa: D107
                 usage: int = GL_DYNAMIC_DRAW) -> None:
        super().__init__(size, usage)

        self.c_type = c_type
        self._ctypes_size = ctypes.sizeof(c_type)
        number = size // self._ctypes_size
        self.data = (c_type * number)()
        self.data_ptr = ctypes.addressof(self.data)

        self._dirty_min = sys.maxsize
        self._dirty_max = 0
        self._dirty = False

        self.stride = stride
        self.count = count

    def commit(self) -> None:
        """Commits all saved changes to the underlying buffer before drawing.

        Allows submitting multiple changes at once, rather than having to call glBufferSubData for every change.
        """
        if not self._dirty:
            return

        glBindBuffer(GL_ARRAY_BUFFER, self.id)
        size = self._dirty_max - self._dirty_min
        if size > 0:
            if size == self.size:
                glBufferData(GL_ARRAY_BUFFER, self.size, self.data, self.usage)
            else:
                glBufferSubData(GL_ARRAY_BUFFER, self._dirty_min, size, self.data_ptr + self._dirty_min)

            self._dirty_min = sys.maxsize
            self._dirty_max = 0
            self._dirty = False

    @lru_cache(maxsize=None)  # noqa: B019
    def get_region(self, start: int, count: int) -> Array[CTypesDataType]:
        byte_start = self.stride * start  # byte offset
        array_count = self.count * count  # number of values
        ptr_type = ctypes.POINTER(self.c_type * array_count)
        return ctypes.cast(self.data_ptr + byte_start, ptr_type).contents

    def set_region(self, start: int, count: int, data: Sequence[float]) -> None:
        array_start = self.count * start
        array_end = self.count * count + array_start

        self.data[array_start:array_end] = data

        # replicated from self.invalidate_region
        byte_start = self.stride * start
        byte_end = byte_start + self.stride * count
        # As of Python 3.11, this is faster than min/max:
        if byte_start < self._dirty_min:
            self._dirty_min = byte_start
        if byte_end > self._dirty_max:
            self._dirty_max = byte_end
        self._dirty = True

    def resize(self, size: int) -> None:
        # size is the allocator size * attribute.stride
        number = size // ctypes.sizeof(self.c_type)
        data = (self.c_type * number)()
        ctypes.memmove(data, self.data, min(size, self.size))
        self.data = data
        self.data_ptr = ctypes.addressof(data)
        self.size = size

        # Set the dirty range to be the entire buffer.
        self._dirty_min = 0
        self._dirty_max = self.size
        self._dirty = True

        self.get_region.cache_clear()

    def invalidate(self) -> None:
        super().invalidate()
        self._dirty = True

    def invalidate_region(self, start: int, count: int) -> None:
        byte_start = self.stride * start
        byte_end = byte_start + self.stride * count
        # As of Python 3.11, this is faster than min/max:
        if byte_start < self._dirty_min:
            self._dirty_min = byte_start
        if byte_end > self._dirty_max:
            self._dirty_max = byte_end
        self._dirty = True


class AttributeBufferObject(BackedBufferObject):
    """A backed buffer used for Shader Program attributes."""

    def __init__(self, size: int, attribute: Attribute) -> None:  # noqa: D107
        # size is the allocator size * attribute.stride (buffer size)
        super().__init__(size, attribute.c_type, attribute.stride, attribute.count)


class IndexedBufferObject(BackedBufferObject):
    """A backed buffer used for indices."""

    def __init__(self, size: int, c_type: CTypesDataType, stride: int, count: int,  # noqa: D107
                 usage: int = GL_DYNAMIC_DRAW) -> None:
        super().__init__(size, c_type, stride, count, usage)


class PersistentBufferObject(AbstractBuffer):
    """A persistently mapped buffer.

    Available in OpenGL 4.3+ contexts. Persistently mapped buffers
    are mapped one time on creation, and can be updated at any time
    without the need to map or unmap.
    """

    def __init__(self, size, attribute, vao):
        # TODO: Persistent buffers cannot be resized. A new buffer is created, and the
        #       data is copied over. Therefore, unlike other buffers, they currently
        #       require s reference to an attribute so the attribute pointer can be reset
        #       on resize calls. This can be reevaluated for a better solution.

        self.size = size
        self.attribute = attribute
        self.attribute_stride = attribute.stride
        self.attribute_count = attribute.count
        self.attribute_ctype = attribute.c_type
        self.vao = vao

        self._context = pyglet.gl.current_context

        buffer_id = GLuint()
        glGenBuffers(1, buffer_id)
        self.id = buffer_id.value
        glBindBuffer(GL_ARRAY_BUFFER, self.id)

        self.flags = GL_MAP_WRITE_BIT | GL_MAP_PERSISTENT_BIT | GL_MAP_COHERENT_BIT
        data = (GLubyte * size)()
        glBufferStorage(GL_ARRAY_BUFFER, size, data, self.flags)

        # size is the allocator size * attribute.stride
        number = size // attribute.element_size
        ptr = ctypes.POINTER(attribute.c_type * number)
        self.data = ctypes.cast(glMapBufferRange(GL_ARRAY_BUFFER, 0, size, self.flags), ptr).contents

    def set_data(self, data: Sequence[int] | CTypesPointer) -> None:
        raise NotImplementedError("Not yet implemented")

    def set_data_region(self, data: Sequence[int] | CTypesPointer, start: int, length: int) -> None:
        raise NotImplementedError("Not yet implemented")

    def bind(self, target=GL_ARRAY_BUFFER):
        glBindBuffer(target, self.id)

    def unbind(self):
        glBindBuffer(GL_ARRAY_BUFFER, 0)

    def map(self) -> CTypesPointer[ctypes.c_ubyte]:
        raise NotImplementedError("PersistentBufferObjects are always mapped.")

    def map_range(self, start, size, ptr_type, flags=GL_MAP_WRITE_BIT):
        raise NotImplementedError("PersistentBufferObjects are always mapped.")

    def unmap(self) -> None:
        raise NotImplementedError("PersistentBufferObjects cannot be unmapped.")

    def delete(self) -> None:
        glDeleteBuffers(1, GLuint(self.id))
        self.id = None

    @lru_cache(maxsize=None)
    def get_region(self, start, count):
        byte_start = self.attribute_stride * start  # byte offset
        array_count = self.attribute_count * count  # number of values

        ptr_type = ctypes.POINTER(self.attribute_ctype * array_count)
        return ctypes.cast(ctypes.addressof(self.data) + byte_start, ptr_type).contents

    def set_region(self, start, count, data):
        array_start = self.attribute_count * start
        array_end = self.attribute_count * count + array_start
        self.data[array_start:array_end] = data

    def resize(self, size):
        # Create temporary copy of current data
        temp = (GLubyte * size)()
        ctypes.memmove(temp, self.data, min(size, self.size))
        glDeleteBuffers(1, GLuint(self.id))

        # Generate new buffer
        buffer_id = GLuint()
        glGenBuffers(1, buffer_id)
        self.id = buffer_id.value

        # Link attributes to new buffer:
        self.vao.bind()
        self.bind()
        self.attribute.enable()
        self.attribute.set_pointer(self.ptr)

        # Initialize the new buffer with the old data, and map it:
        glBufferStorage(GL_ARRAY_BUFFER, size, temp, self.flags)

        ptr_type = self.attribute.c_type * (size // self.attribute.element_size)
        self.data = self.map_range(0, size, ctypes.POINTER(ptr_type), self.flags)

        self.size = size
        self.get_region.cache_clear()

    def sub_data(self):
        # Not necessary with persistent mapping
        pass

    def invalidate(self):
        # Not necessary with persistent mapping
        pass

    def invalidate_region(self, start, count):
        # Not necessary with persistent mapping
        pass



from io import BytesIO
import pytest

from pyglet import gl, image

from ...annotations import Platform, require_platform, require_gl_extension
from ...base.event_loop import EventLoopFixture


class ImageTestFixture(EventLoopFixture):
    def __init__(self, request, test_data):
        super().__init__(request)
        self.test_data = test_data

        self.show_checkerboard = True
        self.show_triangle_left = False
        self.show_text = True
        self.left_texture = None
        self.right_texture = None

        self.checkerboard = image.create(32, 32, image.CheckerImagePattern())

    def on_draw(self):
        # Do not call super class draw, we need to split the clearing and the drawing
        # the text box.
        self.clear()
        self.draw_checkerboard()
        self.draw_left()
        self.draw_triangle_left()
        self.draw_right()
        if self.show_text:
            self.draw_text()

    def draw_checkerboard(self):
        if self.show_checkerboard:
            gl.glPushMatrix()
            gl.glScalef(self.window.width/float(self.checkerboard.width),
                        self.window.height/float(self.checkerboard.height),
                        1.)
            gl.glMatrixMode(gl.GL_TEXTURE)
            gl.glPushMatrix()
            gl.glScalef(self.window.width/float(self.checkerboard.width),
                        self.window.height/float(self.checkerboard.height),
                        1.)
            gl.glMatrixMode(gl.GL_MODELVIEW)
            self.checkerboard.blit(0, 0, 0)
            gl.glMatrixMode(gl.GL_TEXTURE)
            gl.glPopMatrix()
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glPopMatrix()

    def draw_left(self):
        if self.left_texture:
            self.left_texture.blit(
                self.window.width // 4 - self.left_texture.width // 2,
                (self.window.height - self.left_texture.height) // 2,
                0)

    def draw_right(self):
        if self.right_texture:
            x = self.window.width * 3 // 4 - self.right_texture.width // 2
            x = max((x, self.window.width // 2))
            self.right_texture.blit(
                x,
                (self.window.height - self.right_texture.height) // 2,
                0)

    def load_left(self, image_file, decoder=None):
        self.left_texture = image.load(image_file, decoder=decoder).get_texture()

    def copy_left_to_right(self, encoder=None):
        buf = BytesIO()
        self.left_texture.save("file.png",
                               buf,
                               encoder=encoder)
        buf.seek(0)
        self.right_texture = image.load("file.png", buf).get_texture()

    def enable_alpha(self):
        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

    def load_right_arb(self, image_file, pixel_format):
        img = image.load(image_file)
        img.format = pixel_format
        img.get_data()  # forces conversion
        self.right_texture = img.get_texture()

    def draw_triangle_left(self):
        if self.show_triangle_left:
            w = 200
            h = 200
            x = self.window.width // 4 - w // 2
            y = (self.window.height - h) // 2

            gl.glEnable(gl.GL_DEPTH_TEST)
            gl.glBegin(gl.GL_TRIANGLES)
            gl.glColor4f(1, 0, 0, 1)
            gl.glVertex3f(x, y, -1)
            gl.glColor4f(0, 1, 0, 1)
            gl.glVertex3f(x+w, y, 0)
            gl.glColor4f(0, 0, 1, 1)
            gl.glVertex3f(x, y+h, 1)
            gl.glEnd()
            gl.glDisable(gl.GL_DEPTH_TEST)
            gl.glColor4f(1, 1, 1, 1)

    def copy_color_buffer(self):
        self.right_texture = \
                image.get_buffer_manager().get_color_buffer().get_texture()

    def save_and_load_color_buffer(self):
        stream = BytesIO()
        image.get_buffer_manager().get_color_buffer().save('buffer.png', stream)
        stream.seek(0)
        self.right_texture = image.load('buffer.png', stream)

    def save_and_load_depth_buffer(self):
        stream = BytesIO()
        image.get_buffer_manager().get_depth_buffer().save('buffer.png', stream)
        stream.seek(0)
        self.right_texture = image.load('buffer.png', stream)


    def test_image_loading(self, decoder, image_name):
        """Test loading images."""
        self.create_window(width=800, height=600)
        self.load_left(self.test_data.get_file("images", image_name), decoder)
        self.enable_alpha()
        self.ask_question(
                "Do you see the {} image on a checkerboard background?".format(image_name)
                )

    def test_image_saving(self, encoder, image_name):
        """Test saving images."""
        self.create_window(width=800, height=600)
        self.load_left(self.test_data.get_file("images", image_name))
        self.copy_left_to_right(encoder)
        self.enable_alpha()
        self.ask_question(
                "Do you see the {} image twice on a checkerboard background?".format(image_name)
                )

@pytest.fixture
def image_test(request, test_data):
    return ImageTestFixture(request, test_data)


bmp_images = ['rgb_16bpp.bmp', 'rgb_1bpp.bmp', 'rgb_24bpp.bmp', 'rgb_32bpp.bmp', 'rgb_4bpp.bmp',
              'rgb_8bpp.bmp', 'rgba_32bpp.bmp']
dds_images = ['rgba_dxt1.dds', 'rgba_dxt3.dds', 'rgba_dxt5.dds', 'rgb_dxt1.dds']
png_images = ['la.png', 'l.png', 'rgba.png', 'rgb.png']
pypng_images = png_images + ['rgb_8bpp.png', 'rgb_8bpp_trans.png']
gif_images = ['8bpp.gif']


def test_checkerboard(image_test):
    """Test that the checkerboard pattern looks correct."""
    image_test.create_window()
    image_test.ask_question(
            "Do you see a checkboard pattern in two levels of grey?"
            )

@pytest.mark.parametrize('image_name', bmp_images)
def test_bmp_loading(image_test, image_name):
    """Test loading BMP images."""
    from pyglet.image.codecs.bmp import BMPImageDecoder
    image_test.test_image_loading(BMPImageDecoder(), image_name)


@pytest.mark.parametrize('image_name', dds_images)
def test_dds_loading(image_test, image_name):
    """Test loading DDS images."""
    from pyglet.image.codecs.dds import DDSImageDecoder
    image_test.test_image_loading(DDSImageDecoder(), image_name)


@pytest.mark.parametrize('image_name', png_images)
def test_pil_loading(image_test, image_name):
    """Test loading PNG images using PIL"""
    try:
        from PIL import Image
    except ImportError:
        pytest.skip('PIL not available')
    from pyglet.image.codecs.pil import PILImageDecoder
    image_test.test_image_loading(PILImageDecoder(), image_name)


@pytest.mark.parametrize('image_name', png_images + gif_images)
@require_platform(Platform.LINUX)
def test_gdkpixbuf2_loading(image_test, image_name):
    """Test loading PNG images using Linux specific GdkPixbuf2."""
    from pyglet.image.codecs.gdkpixbuf2 import GdkPixbuf2ImageDecoder
    image_test.test_image_loading(GdkPixbuf2ImageDecoder(), image_name)


@pytest.mark.parametrize('image_name', png_images)
@require_platform(Platform.WINDOWS)
def test_gdiplus_loading(image_test, image_name):
    """Test loading PNG images using Windows specific GDI+."""
    from pyglet.image.codecs.gdiplus import GDIPlusDecoder
    image_test.test_image_loading(GDIPlusDecoder(), image_name)


@pytest.mark.parametrize('image_name', png_images)
@require_platform(Platform.OSX)
def test_quartz_loading(image_test, image_name):
    """Test loading PNG images using OSX specific Quartz."""
    from pyglet.image.codecs.quartz import QuartzImageDecoder
    image_test.test_image_loading(QuartzImageDecoder(), image_name)


@pytest.mark.parametrize('image_name', pypng_images)
def test_pypng_loading(image_test, image_name):
    """Test loading PNG images using PyPNG."""
    from pyglet.image.codecs.png import PNGImageDecoder
    image_test.test_image_loading(PNGImageDecoder(), image_name)


@pytest.mark.parametrize('image_name', png_images)
def test_pil_saving(image_test, image_name):
    """Test saving images using PIL."""
    try:
        from PIL import Image
    except ImportError:
        pytest.skip('PIL not available')
    from pyglet.image.codecs.pil import PILImageEncoder
    image_test.test_image_saving(PILImageEncoder(), image_name)


@pytest.mark.parametrize('image_name', png_images)
def test_pypng_saving(image_test, image_name):
    """Test saving images using PyPNG."""
    from pyglet.image.codecs.png import PNGImageEncoder
    image_test.test_image_saving(PNGImageEncoder(), image_name)


@pytest.mark.parametrize('image_name', ['rgb.png', 'rgba.png'])
@require_gl_extension('GL_ARB_imaging')
def test_arb(image_test, image_name):
    """Test swapping color channels using the ARB imaging extension."""
    image_test.create_window()
    image_test.load_left(image_test.test_data.get_file('images', image_name))
    image_test.load_right_arb(image_test.test_data.get_file('images', image_name), 'GRB')
    image_test.ask_question(
            "In the right image red and green should be swapped."
            )


def test_buffer_copy(image_test):
    """Test colour buffer copy to texture.

    A scene consisting of a single coloured triangle will be rendered.  The
    colour buffer will then be saved to a stream and loaded as a texture.

    You might see the original scene first shortly before the
    buffer image appears (because retrieving and saving the image is a slow
    operation).
    """
    image_test.create_window(width=800, height=600)
    image_test.show_triangle_left = True
    image_test.show_text = False
    image_test.show_checkerboard = False

    def step(dt):
        image_test.copy_color_buffer()
        image_test.show_text = True
        return True
    image_test.schedule_once(step)

    image_test.ask_question(
            'You should see the same coloured triangle left and right.'
            )

def test_buffer_saving(image_test):
    """Test colour buffer save.

    A scene consisting of a single coloured triangle will be rendered.  The
    colour buffer will then be saved to a stream and loaded as a texture.

    You might see the original scene first shortly before the
    buffer image appears (because retrieving and saving the image is a slow
    operation).
    """
    image_test.create_window(width=800, height=600)
    image_test.show_triangle_left = True
    image_test.show_text = False
    image_test.show_checkerboard = False

    def step(dt):
        image_test.save_and_load_color_buffer()
        image_test.show_text = True
        return True
    image_test.schedule_once(step)

    image_test.ask_question(
            'You should see the same coloured triangle left and right.'
            )

def test_depth_buffer_saving(image_test):
    """Test depth buffer save.

    A scene consisting of a single coloured triangle will be rendered.  The
    depth buffer will then be saved to a stream and loaded as a texture.

    You might see the original scene first for up to several seconds before the
    depth buffer image appears (because retrieving and saving the image is
    a slow operation).
    """
    image_test.create_window(width=800, height=600)
    image_test.show_triangle_left = True
    image_test.show_text = False
    image_test.show_checkerboard = False

    def step(dt):
        image_test.save_and_load_depth_buffer()
        image_test.show_text = True
        return True
    image_test.schedule_once(step)

    image_test.ask_question(
            'You should see a coloured triangle left and its depth buffer right. '
            'The bottom-left corner is lightest, the bottom-right is darker and '
            'the top corner is darkest (corresponding the depth of the triangle.'
            )




"""Tests clock timing between frames and estimations
of frames per second.
"""
import time
import pytest

from pyglet import clock

from ..annotations import skip_if_continuous_integration


def sleep(seconds):
    """Busy sleep on the CPU which is very precise"""
    start = time.perf_counter()
    while time.perf_counter() - start < seconds:
        pass


@pytest.fixture
def newclock():
    clock.set_default(clock.Clock())
    yield clock


def test_first_tick_is_delta_zero(newclock):
    """
    Tests that the first tick is dt = 0.
    """
    dt = newclock.tick()
    assert dt == 0


def test_start_at_zero_fps(newclock):
    """
    Tests that the default clock starts
    with zero fps.
    """
    fps = newclock.get_frequency()
    assert fps == 0


@skip_if_continuous_integration()
def test_elapsed_time_between_tick():
    """
    Test that the tick function returns the correct elapsed
    time between frames, in seconds.

    Because we are measuring time differences, we
    expect a small error (1%) from the expected value.
    """
    sleep_time = 0.2

    # initialize internal counter
    clock.tick()

    # test between initialization and first tick
    sleep(sleep_time)
    delta_time_1 = clock.tick()

    # test between non-initialization tick and next tick
    sleep(sleep_time)
    delta_time_2 = clock.tick()

    assert delta_time_1 == pytest.approx(sleep_time, rel=0.01*sleep_time)
    assert delta_time_2 == pytest.approx(sleep_time, rel=0.01*sleep_time)


@skip_if_continuous_integration()
def test_compute_fps():
    """
    Test that the clock computes a reasonable value of
    frames per second when simulated for 120 ticks at 60 frames per second.

    Because sleep is not very precise and fps are unbounded, we
    expect a moderate error (10%) from the expected value.
    """
    ticks = 120  # for averaging
    expected_fps = 60
    seconds_per_tick = 1./expected_fps

    for i in range(ticks):
        time.sleep(seconds_per_tick)
        clock.tick()
    computed_fps = clock.get_frequency()

    assert computed_fps == pytest.approx(expected_fps, rel=0.1*expected_fps)




import pytest
from tests import mock
import time

import pyglet
_debug = False
pyglet.options['debug_media'] = _debug

from pyglet.media import Player
from pyglet.media.synthesis import Silence
from .mock_player import MockPlayer


class PlayerTest(MockPlayer, Player):
    pass


@pytest.fixture
def player(event_loop):
    return PlayerTest(event_loop)


class SilentTestSource(Silence):
    def __init__(self, duration, frequency=440, sample_rate=44800, envelope=None):
        super().__init__(duration, frequency, sample_rate, envelope)
        self.bytes_read = 0

    def get_audio_data(self, nbytes):
        data = super().get_audio_data(nbytes)
        if data is not None:
            self.bytes_read += data.length
        return data

    @property
    def max_offset(self):
        return self._max_offset


def test_player_play(player):
    source = SilentTestSource(.1)
    player.queue(source)

    player.play()
    player.wait_for_all_events(
        1.,
        'on_eos',
        'on_player_eos'
    )
    assert source.bytes_read == source.max_offset, 'Source not fully played'


def test_player_play_multiple(player):
    sources = (SilentTestSource(.1), SilentTestSource(.1))
    for source in sources:
        player.queue(source)

    player.play()
    player.wait_for_all_events(
        1.,
        'on_eos',
        'on_player_next_source',
        'on_eos',
        'on_player_eos'
    )
    for source in sources:
        assert source.bytes_read == source.max_offset, 'Source not fully played'


def test_multiple_fire_and_forget_players():
    """
    Test an issue where the driver crashed when starting multiple players, but not
    keeping a reference to these players.
    """
    for _ in range(10):
        Silence(1).play()
    time.sleep(1)


def test_player_silent_audio_driver(player):
    with mock.patch('pyglet.media.player.get_audio_driver') as get_audio_driver_mock:
        get_audio_driver_mock.return_value = None
        source = SilentTestSource(.1)
        player.queue(source)
        player.play()

        player.wait_for_all_events(
            1.,
            'on_eos',
            'on_player_eos')



from ctypes import *

import pyglet
import pyglet.util


__all__ = ['link_EGL']

egl_lib = pyglet.lib.load_library('EGL')

# Look for eglGetProcAddress
eglGetProcAddress = getattr(egl_lib, 'eglGetProcAddress')
eglGetProcAddress.restype = POINTER(CFUNCTYPE(None))
eglGetProcAddress.argtypes = [POINTER(c_ubyte)]


def link_EGL(name, restype, argtypes, requires=None, suggestions=None):
    try:
        func = getattr(egl_lib, name)
        func.restype = restype
        func.argtypes = argtypes
        return func
    except AttributeError:
        bname = cast(pointer(create_string_buffer(pyglet.util.asbytes(name))), POINTER(c_ubyte))
        addr = eglGetProcAddress(bname)
        if addr:
            ftype = CFUNCTYPE(*((restype,) + tuple(argtypes)))
            func = cast(addr, ftype)
            return func

    return pyglet.gl.lib.missing_function(name, requires, suggestions)



"""Events for :py:mod:`pyglet.window`.

See :py:class:`~pyglet.window.Window` for a description of the window event types.
"""
from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from pyglet.window import key, mouse

if TYPE_CHECKING:
    from io import StringIO


class WindowEventLogger:
    """Print all events to a file.

    When this event handler is added to a window it prints out all events
    and their parameters; useful for debugging or discovering which events
    you need to handle.

    Example::

        win = window.Window()
        event_logger = WindowEventLogger()
        win.push_handlers(event_logger)

    """

    def __init__(self, logfile: StringIO | None = None) -> None:
        """Create an event logger which writes to ``logfile``.

        Args:
            logfile:
                The file to write to.  If unspecified, stdout will be used.
        """
        if logfile is None:
            logfile = sys.stdout
        self.file = logfile

    def on_key_press(self, symbol: int, modifiers: int) -> None:
        print(f'on_key_press(symbol={key.symbol_string(symbol)}, modifiers={key.modifiers_string(modifiers)})',
              file=self.file)

    def on_key_release(self, symbol: int, modifiers: int) -> None:
        print(f'on_key_release(symbol={key.symbol_string(symbol)}, modifiers={key.modifiers_string(modifiers)})',
              file=self.file)

    def on_text(self, text: str) -> None:
        print(f'on_text(text={text!r})', file=self.file)

    def on_text_motion(self, motion: int) -> None:
        print(f'on_text_motion(motion={key.motion_string(motion)})', file=self.file)

    def on_text_motion_select(self, motion: int) -> None:
        print(f'on_text_motion_select(motion={key.motion_string(motion)})', file=self.file)

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> None:
        print(f'on_mouse_motion(x={x}, y={y}, dx={dx}, dy={dy})', file=self.file)

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int) -> None:
        print(
            f'on_mouse_drag(x={x}, y={y}, dx={dx}, dy={dy}, buttons={mouse.buttons_string(buttons)}, modifiers='
            f'{key.modifiers_string(modifiers)})',
            file=self.file)

    def on_mouse_press(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        print(
            f'on_mouse_press(x={x}, y={y}, button={mouse.buttons_string(buttons)!r}, modifiers='
            f'{key.modifiers_string(modifiers)})',
            file=self.file)

    def on_mouse_release(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        print(
            f'on_mouse_release(x={x}, y={y}, button={mouse.buttons_string(buttons)!r}, modifiers='
            f'{key.modifiers_string(modifiers)})',
            file=self.file)

    def on_mouse_scroll(self, x: int, y: int, scroll_x: float, scroll_y: float) -> None:
        print(f'on_mouse_scroll(x={x}, y={y}, scroll_x={scroll_x}, scroll_y={scroll_y})', file=self.file)

    def on_close(self) -> None:
        print('on_close()', file=self.file)

    def on_mouse_enter(self, x: int, y: int) -> None:
        print('on_mouse_enter(x=%d, y=%d)' % (x, y), file=self.file)

    def on_mouse_leave(self, x: int, y: int) -> None:
        print('on_mouse_leave(x=%d, y=%d)' % (x, y), file=self.file)

    def on_expose(self) -> None:
        print('on_expose()', file=self.file)

    def on_resize(self, width: int, height: int) -> None:
        print(f'on_resize(width={width}, height={height})', file=self.file)

    def on_move(self, x: int, y: int) -> None:
        print(f'on_move(x={x}, y={y})', file=self.file)

    def on_activate(self) -> None:
        print('on_activate()', file=self.file)

    def on_deactivate(self) -> None:
        print('on_deactivate()', file=self.file)

    def on_show(self) -> None:
        print('on_show()', file=self.file)

    def on_hide(self) -> None:
        print('on_hide()', file=self.file)

    def on_context_lost(self) -> None:
        print('on_context_lost()', file=self.file)

    def on_context_state_lost(self) -> None:
        print('on_context_state_lost()', file=self.file)

    def on_draw(self) -> None:
        print('on_draw()', file=self.file)



"""Group multiple small images into larger Textures.

This module provides classes to efficiently pack small images into larger Textures.
This can have major performance benefits when dealiing with a large number of images.

:py:class:`~pyglet.image.atlas.TextureAtlas` maintains one texture; :py:class:`TextureBin`
manages a collection of atlases of a given size. :py:class:`TextureArrayBin` works similarly
except for :py:class:`~pyglet.image.TextureArray`s instead of altases.

This module is used internally by the :py:mod:`pyglet.resource` module.

Example usage::

    # Load images from disk
    car_image = pyglet.image.load('car.png')
    boat_image = pyglet.image.load('boat.png')

    # Pack these images into one or more textures
    bin = TextureBin()
    car_texture = bin.add(car_image)
    boat_texture = bin.add(boat_image)

The result of :py:meth:`TextureBin.add` is a :py:class:`TextureRegion`
containing the image. Once added, an image cannot be removed from a bin (or an 
atlas); nor can a list of images be obtained from a given bin or atlas -- it is 
the application's responsibility to keep track of the regions returned by the
``add`` methods.

.. versionadded:: 1.1
"""
from __future__ import annotations

import pyglet

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyglet.image import AbstractImage, ImageData, TextureRegion, TextureArrayRegion


class AllocatorException(Exception):
    """The allocator does not have sufficient free space for the requested
    image size."""
    pass


class _Strip:
    __slots__ = 'x', 'y', 'max_height', 'y2'

    def __init__(self, y: int, max_height: int) -> None:
        self.x = 0
        self.y = y
        self.max_height = max_height
        self.y2 = y

    def add(self, width: int, height: int) -> tuple[int, int]:
        assert width > 0 and height > 0
        assert height <= self.max_height

        x, y = self.x, self.y
        self.x += width
        self.y2 = max(self.y + height, self.y2)
        return x, y

    def compact(self) -> None:
        self.max_height = self.y2 - self.y


class Allocator:
    """Rectangular area allocation algorithm.

    An ``Allocator`` is initialized with a specified ``width`` and ``height``.
    The :py:meth:`~alloc` method can then be called to retrieve free regions
    of that area (and protect them from future allocations).

    ``Allocator`` uses a fairly simple strips-based algorithm. It performs
    best when rectangles are allocated of the same size, or in decreasing
    height order.
    """
    __slots__ = 'width', 'height', 'strips', 'used_area'

    def __init__(self, width: int, height: int) -> None:
        """Create an ``Allocator`` of the given size."""
        assert width > 0 and height > 0
        self.width = width
        self.height = height
        self.strips = [_Strip(0, height)]
        self.used_area = 0

    def alloc(self, width: int, height: int) -> tuple[int, int]:
        """Get the position of a free area in the allocator of the given size

        If a suitable position can be found for the requested size, the position
        will be marked as in-use and returned. (The same position will never be
        returned twice).
        If there is not enough room to fit the given area, ``AllocatorException``
        is raised.
        """
        for strip in self.strips:
            if self.width - strip.x >= width and strip.max_height >= height:
                self.used_area += width * height
                return strip.add(width, height)

        strip = self.strips[-1]

        if self.width >= width and self.height - strip.y2 >= height:
            self.used_area += width * height
            strip.compact()
            newstrip = _Strip(strip.y2, self.height - strip.y2)
            self.strips.append(newstrip)
            return newstrip.add(width, height)

        raise AllocatorException(f"No more space in {self} for box {width}x{height}")

    def get_usage(self) -> float:
        """Get the fraction of area already allocated.

        This method is useful for debugging and profiling only.
        """
        return self.used_area / float(self.width * self.height)

    def get_fragmentation(self) -> float:
        """Get the fraction of area that's unlikely to ever be used, based on
        current allocation behaviour.

        This method is useful for debugging and profiling only.
        """
        # The total unused area in each compacted strip is summed.
        if not self.strips:
            return 0.0
        possible_area = self.strips[-1].y2 * self.width
        return 1.0 - self.used_area / possible_area


class TextureAtlas:
    """A large Texture made up of multiple smaller images.

    A TextureAtlas is one maximally sized Texture which is made up of
    multiple smaller Images. This can improve rendering performance
    by allowing all the Images to be drawn together with a single
    Texture bind, rather than multiple tiny Texture binds per draw.

    When creating a TextureAtlas instance, a new :py:class:`~pyglet.image.Texture`
    object will be created at the requested size. If the maximum texture
    size supported by the OpenGL driver is less than requested, the
    smaller of the two will be used.
    """

    def __init__(self, width: int = 2048, height: int = 2048) -> None:
        """Create a Texture Atlas of the given size."""
        max_texture_size = pyglet.image.get_max_texture_size()
        width = min(width, max_texture_size)
        height = min(height, max_texture_size)

        self.texture = pyglet.image.Texture.create(width, height)
        self.allocator = Allocator(width, height)

    def add(self, img: ImageData, border: int = 0) -> TextureRegion:
        """Add ImageData to the atlas.

        Given :py:class:`~pyglet.image.ImageData`, add it to the Atlas and
        return a new :py:class:`~pyglet.image.TextureRegion` object. An
        optional ``border`` argument can be passed, which will leave the
        specified number of blank pixels around the added ImageData. This
        can be useful in certain situations and blend modes, where
        neighboring pixel data can "bleed" into the edges.

        This method will fail if the given image cannot be transferred directly
        to a texture (for example, if it is not an instance of ImageData, such
        as another Texture). ``AllocatorException`` will be raised if there is
        no room in the atlas for the image.
        """
        x, y = self.allocator.alloc(img.width + border * 2, img.height + border * 2)
        self.texture.blit_into(img, x + border, y + border, 0)
        return self.texture.get_region(x + border, y + border, img.width, img.height)


class TextureBin:
    """Collection of TextureAtlases.

    :py:class:`~pyglet.image.atlas.TextureBin` maintains a collection
    of TextureAtlases, and creates new ones as necessary to accommodate
    adding an unbound number of Images. When one TextureAltas is full,
    a new one is automatically created to fit the next ImageData.
    """

    def __init__(self, texture_width: int = 2048, texture_height: int = 2048) -> None:
        """Create a texture bin for holding atlases of the given size."""
        max_texture_size = pyglet.image.get_max_texture_size()
        self.texture_width = min(texture_width, max_texture_size)
        self.texture_height = min(texture_height, max_texture_size)
        self.atlases = []

    def add(self, img: ImageData | AbstractImage, border: int = 0) -> TextureRegion:
        """Add an image into this texture bin.

        This method calls :py:meth:`~TextureAtlas.add` for the first atlas
        that has room for the image.

        ``AllocatorException`` is raised if the image exceeds the dimensions
        of ``texture_width`` and ``texture_height``.
        """
        for atlas in list(self.atlases):
            try:
                return atlas.add(img, border)
            except AllocatorException:
                # Remove atlases that are no longer useful (so that their textures
                # can later be freed if the images inside them get collected).
                if img.width < 64 and img.height < 64:
                    self.atlases.remove(atlas)

        atlas = TextureAtlas(self.texture_width, self.texture_height)
        self.atlases.append(atlas)
        return atlas.add(img, border)


class TextureArrayBin:
    """Collection of texture arrays.

    :py:class:`~pyglet.image.atlas.TextureArrayBin` maintains a collection of
    texture arrays, and creates new ones as necessary as the depth is exceeded.
    This works similarly to TextureBin, but it manages TextureArrays instead of
    TextureAtlases.
    """

    def __init__(self, texture_width: int = 2048, texture_height: int = 2048, max_depth: int | None = None) -> None:
        max_texture_size = pyglet.image.get_max_texture_size()
        self.max_depth = max_depth or pyglet.image.get_max_array_texture_layers()
        self.texture_width = min(texture_width, max_texture_size)
        self.texture_height = min(texture_height, max_texture_size)
        self.arrays = []

    def add(self, img: ImageData) -> TextureArrayRegion:
        """Add an image into this texture array bin.

        This method calls :py:meth:`~pyglet.image.TextureArray.add` for the first
        array that has room for the image.

        ``TextureArraySizeExceeded`` is raised if the image exceeds the dimensions of
        ``texture_width`` and ``texture_height``.
        """
        try:
            array = self.arrays[-1]
            return array.add(img)
        except pyglet.image.TextureArrayDepthExceeded:
            pass
        except IndexError:
            pass

        array = pyglet.image.TextureArray.create(self.texture_width, self.texture_height, max_depth=self.max_depth)
        self.arrays.append(array)
        return array.add(img)



"""A simple demonstration of the HTMLLabel class, as it might be used on a
help or introductory screen.
"""

import os
import pyglet

html = """
<h1>HTML labels in pyglet</h1>

<p align="center"><img src="pyglet.png" /></p>

<p>HTML labels are a simple way to add formatted text to your application.
Different <font face="Helvetica,Arial" size=+2>fonts</font>, <em>styles</em>,
<u>underlines</u>, and <font color=maroon>colours</font> are supported.

<p>This window has been made resizable; text will reflow to fit the new size.
"""

window = pyglet.window.Window(resizable=True)
location = pyglet.resource.FileLocation(os.path.dirname(__file__))
label = pyglet.text.HTMLLabel(html, location=location,
                              width=window.width,
                              multiline=True, anchor_y='center')


@window.event
def on_resize(width, height):
    # Wrap text to the width of the window
    label.width = window.width

    # Keep text vertically centered in the window
    label.y = window.height // 2


@window.event
def on_draw():
    window.clear()
    label.draw()


pyglet.gl.glClearColor(1, 1, 1, 1)
pyglet.app.run()



"""
pytest hooks for interactive test cases.
"""
import inspect
import pytest


def pytest_collection_modifyitems(items, config):
    """Determine whether test should be skipped based on command-line options."""
    sanity = config.getoption('--sanity', False)
    non_interactive = config.getoption('--non-interactive', False)

    remaining = []
    deselected = []
    for item in items:
        if _skip_item(item, sanity, non_interactive):
            deselected.append(item)
        else:
            remaining.append(item)
    if deselected:
        items[:] = remaining
        config.hook.pytest_deselected(items=deselected)


def _skip_item(item, sanity, non_interactive):
    requires_user_action = item.get_closest_marker('requires_user_action')
    requires_user_validation = item.get_closest_marker('requires_user_validation')
    only_interactive = item.get_closest_marker('only_interactive')

    if ((requires_user_action is not None or only_interactive is not None) 
            and (sanity or non_interactive)):
        return True

    if ((requires_user_validation is not None) 
            and (non_interactive)):
        return True

    return False

def pytest_runtest_setup(item):
    # TODO: Remove after migrating tests
    sanity = item.config.getoption('--sanity', False)
    non_interactive = item.config.getoption('--non-interactive', False)
    interactive = not sanity and not non_interactive

    if interactive:
        _show_test_header(item)
    _try_set_class_attribute(item, 'interactive', interactive)
    _try_set_class_attribute(item, 'allow_missing_screenshots', sanity)

def _show_test_header(item):
    print()
    print('='*80)
    print(item.name)
    print(_get_doc(item))
    print('-'*80)

def _try_set_class_attribute(item, name, value):
    if hasattr(item.obj, 'im_class'):
        if hasattr(item.obj.im_class, name):
            setattr(item.obj.im_class, name, value)

def _get_doc(item):
    i = item
    while i is not None:
        if hasattr(i, 'obj') and hasattr(i.obj, '__doc__') and i.obj.__doc__ is not None:
            return inspect.cleandoc(i.obj.__doc__)
        i = i.parent

def pytest_runtest_makereport(item, call):
    if call.when == 'call' and call.excinfo is None:
        _legacy_check_screenshots(item)
        _commit_screenshots(item)

def _legacy_check_screenshots(item):
    # TODO: Remove after migrating all tests
    if hasattr(item, 'obj') and hasattr(item.obj, '__self__'):
        if hasattr(item.obj.__self__, 'check_screenshots'):
            item.obj.__self__.check_screenshots()

def _commit_screenshots(item):
    if hasattr(item.session, 'pending_screenshots'):
        for fixture in item.session.pending_screenshots:
            fixture.commit_screenshots()




from __future__ import annotations

from typing import Sequence

from pyglet.display.headless import HeadlessCanvas

# from pyglet.window import key
# from pyglet.window import mouse
from pyglet.event import EventDispatcher
from pyglet.libs.egl import egl
from pyglet.window import (
    BaseWindow,
    DefaultMouseCursor,  # noqa: F401
    ImageMouseCursor,  # noqa: F401
    MouseCursor,  # noqa: F401
    MouseCursorException,  # noqa: F401
    NoSuchDisplayException,  # noqa: F401
    WindowException,  # noqa: F401
    _PlatformEventHandler,
    _ViewEventHandler,
)

# Platform event data is single item, so use platform event handler directly.
HeadlessEventHandler = _PlatformEventHandler
ViewEventHandler = _ViewEventHandler


class HeadlessWindow(BaseWindow):
    _egl_display_connection = None
    _egl_surface = None

    def _recreate(self, changes: Sequence[str]) -> None:
        pass

    def flip(self) -> None:
        if self.context:
            self.context.flip()

    def switch_to(self) -> None:
        if self.context:
            self.context.set_current()

    def set_caption(self, caption: str) -> None:
        pass

    def set_minimum_size(self, width: int, height: int) -> None:
        pass

    def set_maximum_size(self, width: int, height: int) -> None:
        pass

    def set_size(self, width: int, height: int) -> None:
        pass

    def get_size(self) -> tuple[int, int]:
        return self._width, self._height

    def set_location(self, x: int, y: int) -> None:
        pass

    def get_location(self) -> None:
        pass

    def activate(self) -> None:
        pass

    def set_visible(self, visible: bool = True) -> None:
        pass

    def minimize(self) -> None:
        pass

    def maximize(self) -> None:
        pass

    def set_vsync(self, vsync: bool) -> None:
        pass

    def set_mouse_platform_visible(self, platform_visible: bool | None = None) -> None:
        pass

    def set_exclusive_mouse(self, exclusive: bool = True) -> None:
        pass

    def set_exclusive_keyboard(self, exclusive: bool = True) -> None:
        pass

    def get_system_mouse_cursor(self, name: str) -> None:
        pass

    def dispatch_events(self) -> None:
        while self._event_queue:
            EventDispatcher.dispatch_event(self, *self._event_queue.pop(0))

    def dispatch_pending_events(self) -> None:
        pass

    def _create(self) -> None:
        self._egl_display_connection = self.display._display_connection  # noqa: SLF001

        if not self._egl_surface:
            pbuffer_attribs = (egl.EGL_WIDTH, self._width, egl.EGL_HEIGHT, self._height, egl.EGL_NONE)
            pbuffer_attrib_array = (egl.EGLint * len(pbuffer_attribs))(*pbuffer_attribs)
            self._egl_surface = egl.eglCreatePbufferSurface(self._egl_display_connection,
                                                            self.config._egl_config,  # noqa: SLF001
                                                            pbuffer_attrib_array)

            self.canvas = HeadlessCanvas(self.display, self._egl_surface)

            self.context.attach(self.canvas)

            self.dispatch_event('_on_internal_resize', self._width, self._height)


__all__ = ['HeadlessWindow']
